#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional

from numba import njit, prange
from numpy import (pi, arccos, arctan2, sin, cos, sqrt, sign, copysign, mod, zeros_like, zeros, linspace, floor, arcsin,
                   int, around, nan, full, ndarray)
from scipy.constants import G

HALF_PI = 0.5 * pi
TWO_PI = 2.0 * pi
INV_PI = 1.0 / pi

D_H = 24.
D_M = 60 * D_H
D_S = 60 * D_M

au,   au_e             = 1.496e11, 0.0
msun, msun_e           = 1.9891e30, 0.0
rsun, rsun_e           = 0.5*1.392684e9, 0.0

# Utilities
# =========

cache = False

def epoch(time, zero_epoch, period):
    return around((time - zero_epoch) / period).astype(int)


def fold(time: ndarray, period: float, t0: Optional[float] = 0.0, shift: Optional[float] = 0.0):
    """Folds time over a given period.

    Parameters
    ----------
    time
      An array of time values.

    period
      Folding period.

    origo
      Time of zero phase.

    shift
      Shift in normalized phase [0, 1].

    Returns
    -------
    An array of phase values in the same units as the period.
    """
    phase = ((time - t0) / period + shift + 0.5) % 1.
    return period * (phase - 0.5)


@njit
def mean_anomaly_offset(e, w):
    mean_anomaly_offset = arctan2(sqrt(1.0-e**2) * sin(HALF_PI - w), e + cos(HALF_PI - w))
    mean_anomaly_offset -= e*sin(mean_anomaly_offset)
    return mean_anomaly_offset


@njit
def z_from_ta_s(Ta, a, i, e, w):
    z  = a*(1.0-e**2)/(1.0+e*cos(Ta)) * sqrt(1.0 - sin(w+Ta)**2 * sin(i)**2)
    z *= copysign(1.0, sin(w+Ta))
    return z


@njit(parallel=True)
def z_from_ta_v(Ta, a, i, e, w):
    z  = a*(1.0-e**2)/(1.0+e*cos(Ta)) * sqrt(1.0 - sin(w+Ta)**2 * sin(i)**2)
    z *= sign(1.0, sin(w+Ta))
    return z

@njit
def rclip(v, vmin, vmax):
    return min(max(v, vmin), vmax)


@njit
def iclip(v, vmin, vmax):
    return int(min(max(v, vmin), vmax))


# Mean Anomaly
# ============

@njit(cache=cache)
def mean_anomaly(t, t0, p, e, w):
    offset = mean_anomaly_offset(e, w)
    Ma = mod(TWO_PI * (t - (t0 - offset * p / TWO_PI)) / p, TWO_PI)
    return Ma

@njit(parallel=True)
def mean_anomaly_p(t, t0, p, e, w):
    offset = mean_anomaly_offset(e, w)
    Ma = mod(TWO_PI * (t - (t0 - offset * p / TWO_PI)) / p, TWO_PI)
    return Ma


# Ecccentric anomaly
# ==================

@njit(cache=cache)
def ea_newton_v(t, t0, p, e, w):
    Ma = mean_anomaly(t, t0, p, e, w)
    Ea = Ma.copy()
    for j in range(len(t)):
        err = 0.05
        k = 0
        while abs(err) > 1e-8 and k<1000:
            err   = Ea[j] - e*sin(Ea[j]) - Ma[j]
            Ea[j] = Ea[j] - err/(1.0-e*cos(Ea[j]))
            k += 1
    return Ea

@njit(cache=cache)
def ea_newton_s(t, t0, p, e, w):
    Ma = mean_anomaly(t, t0, p, e, w)
    Ea = Ma
    err = 0.05
    k = 0
    while abs(err) > 1e-8 and k<1000:
        err   = Ea - e*sin(Ea) - Ma
        Ea = Ea - err/(1.0-e*cos(Ea))
        k += 1
    return Ea

@njit(cache=cache)
def ea_iter_s(t, t0, p, e, w):
    Ma = mean_anomaly(t, t0, p, e, w)
    ec = e*sin(Ma)/(1.0 - e*cos(Ma))
    for k in range(15):
        ect = ec
        ec  = e*sin(Ma+ec)
        if (abs(ect-ec) < 1e-4):
            break
    Ea  = Ma + ec
    return Ea

@njit(parallel=True)
def ea_iter_v(t, t0, p, e, w):
    Ma = mean_anomaly(t, t0, p, e, w)
    ec = e*sin(Ma)/(1.0 - e*cos(Ma))
    for j in prange(len(t)):
        for k in range(15):
            ect   = ec[j]
            ec[j] = e*sin(Ma[j]+ec[j])
            if (abs(ect-ec[j]) < 1e-4):
                break
    Ea  = Ma + ec
    return Ea

# True Anomaly
# ============

@njit(parallel=True)
def ta_from_ea_v(Ea, e):
    sta = sqrt(1.0-e**2) * sin(Ea)/(1.0-e*cos(Ea))
    cta = (cos(Ea)-e)/(1.0-e*cos(Ea))
    Ta  = arctan2(sta, cta)
    return Ta

@njit(cache=cache)
def ta_from_ea_s(Ea, e):
    sta = sqrt(1.0-e**2) * sin(Ea)/(1.0-e*cos(Ea))
    cta = (cos(Ea)-e)/(1.0-e*cos(Ea))
    Ta  = arctan2(sta, cta)
    return Ta

@njit(cache=cache)
def ta_newton_s(t, t0, p, e, w):
    return ta_from_ea_s(ea_newton_s(t, t0, p, e, w), e)

@njit
def ta_newton_v(t, t0, p, e, w):
    return ta_from_ea_v(ea_newton_v(t, t0, p, e, w), e)

@njit(cache=cache)
def ta_iter_s(t, t0, p, e, w):
    return ta_from_ea_s(ea_iter_s(t, t0, p, e, w), e)

@njit
def ta_iter_v(t, t0, p, e, w):
    return ta_from_ea_v(ea_iter_v(t, t0, p, e, w), e)

@njit(parallel=True)
def ta_ps3(t, t0, p, e, w):
    Ma = mean_anomaly(t, t0, p, e, w)
    Ta = (Ma + (2.0*e - 0.25*e**3)*sin(Ma)
             + 1.25*e**2*sin(2.0*Ma)
             + 13.0/12.0*e**3*sin(3.0*Ma))
    return Ta

@njit(parallel=True)
def ta_ps5(t, t0, p, e, w):
    Ma = mean_anomaly(t, t0, p, e, w)
    Ta = (Ma + (2.0*e - 0.25*e**3 + 5.0/96.0*e**5) * sin(Ma)
             + (1.25*e**2 - 11.0/24.0*e**4) * sin(2.0*Ma)
             + (13.0/12.0 * e**3 - 43.0/64.0 * e**5) * sin(3.0*Ma)
             + 103.0/96.0 * e**4 * sin(4.0*Ma)
             + 1097.0/960.0 * e**5 * sin(5.0*Ma))
    return Ta

@njit
def ta_from_ma(Ma, e, cache=cache):
    Ta = zeros_like(Ma)
    for j in range(len(Ma)):
        ec = e*sin(Ma[j])/(1.0 - e*cos(Ma[j]))
        for k in range(15):
            ect = ec
            ec  = e*sin(Ma[j]+ec)
            if abs(ect-ec) < 1e-4:
                break
        Ea = Ma[j] + ec
        sta = sqrt(1.0-e**2) * sin(Ea)/(1.0-e*cos(Ea))
        cta = (cos(Ea)-e)/(1.0-e*cos(Ea))
        Ta[j] = arctan2(sta, cta)
    return Ta

@njit
def ta_ip_calculate_table(ne=256, nm=512, cache=cache):
    es = linspace(0, 0.95, ne)
    ms = linspace(0,   pi, nm)
    tae = zeros((ne, nm))
    for i,e in enumerate(es):
        tae[i,:]  = ta_from_ma(ms, e)
        tae[i,:] -= ms
    return tae, es, ms

@njit(parallel=True)
def ta_ip(t, t0, p, e, w, es, ms, tae):
    ne = es.size
    nm = ms.size
    de = es[1] - es[0]
    dm = ms[1] - ms[0]

    ie = iclip(         floor(e/de),      0,   ne-1)
    ae = rclip((e - de*(ie-1)) / de,    0.0,    1.0)
    tae2 = tae[ie:ie+2,:]

    Ma = mean_anomaly(t, t0, p, e, w)
    Ta = zeros_like(Ma)

    for i in range(len(t)):
        if Ma[i] < pi:
            im = iclip(floor(Ma[i]/dm), 0, nm-1)
            am = rclip((Ma[i] - dm*(im-1)) / dm, 0.0, 1.0)
            s = 1.0
        else:
            im = iclip(floor((TWO_PI - Ma[i])/dm), 0, nm-1)
            am = rclip((TWO_PI - (Ma[i] - dm*(im-1))) / dm, 0.0, 1.0)
            s = -1.0

        Ta[i] = ( tae2[0,im  ]*(1.0-ae)*(1.0-am)
                + tae2[1,im  ]*     ae *(1.0-am)
                + tae2[0,im+1]*(1.0-ae)*     am
                + tae2[1,im+1]*     ae *     am  )
        Ta[i] = Ma[i] + s * Ta[i]

        if (Ta[i] < 0.0):
            Ta[i] = Ta[i] + TWO_PI

    return Ta

# Projected distance Z
# ====================
# These functions calculate the projected distance (z) using different ways to calculate the
# tue anomaly (Ta). The functions have different versions optimized for different use-cases
#
#  - z_*_s  : scalar time
#  - z_*_v  : vector time
#  - z_*_p  : vector time, parallelized and usually fastest
#  - z_*_mp : vector time and two-dimensional parameter array, can be faster than the others
#

# Z: Newton's method
# ------------------


@njit(cache=cache)
def z_circular(t, pv):
    t0, p, a, i, e, w = pv
    cosph = cos(TWO_PI*(t-t0)/p)
    z = sign(cosph) * a*sqrt(1.0 - cosph*cosph*sin(i)**2)
    return z


@njit(fastmath=True, parallel=False)
def z_ip_v(t, t0, p, a, i, e, w, es, ms, tae):
    if e < 0. or e >= es[-1]:
        return full(t.size, nan)

    Ma = mean_anomaly(t, t0, p, e, w)
    if e < 0.01:
        Ta = Ma
    else:
        de = es[1] - es[0]
        dm = ms[1] - ms[0]

        ie = int(floor(e / de))
        ae = (e - de * ie) / de
        tae2 = tae[ie:ie + 2, :]

        Ta = zeros_like(Ma)

        for j in range(len(t)):
            if Ma[j] < pi:
                x = Ma[j]
                s = 1.
            else:
                x = TWO_PI - Ma[j]
                s = -1.

            im = int(floor(x / dm))
            am = (x - im * dm) / dm

            Ta[j] = (tae2[0, im] * (1.0 - ae) * (1.0 - am)
                     + tae2[1, im] * ae * (1.0 - am)
                     + tae2[0, im + 1] * (1.0 - ae) * am
                     + tae2[1, im + 1] * ae * am)
            Ta[j] = Ma[j] + s * Ta[j]

            if (Ta[j] < 0.0):
                Ta[j] = Ta[j] + TWO_PI

    z = a * (1.0 - e ** 2) / (1.0 + e * cos(Ta)) * sqrt(1.0 - sin(w + Ta) ** 2 * sin(i) ** 2)
    z *= copysign(1.0, sin(w + Ta))
    return z


@njit(fastmath=True)
def z_ip_s(t, t0, p, a, i, e, w, es, ms, tae):
    if e < 0. or e >= es[-1]:
        return nan

    Ma = mean_anomaly(t, t0, p, e, w)
    if e < 0.01:
        Ta = Ma
    else:
        de = es[1] - es[0]
        dm = ms[1] - ms[0]

        ie = int(floor(e / de))
        ae = (e - de * ie) / de

        if Ma < pi:
            x = Ma
            s = 1.
        else:
            x = TWO_PI - Ma
            s = -1.

        im = int(floor(x / dm))
        am = (x - im * dm) / dm

        Ta = (tae[ie, im] * (1.0 - ae) * (1.0 - am)
              + tae[ie + 1, im] * ae * (1.0 - am)
              + tae[ie, im + 1] * (1.0 - ae) * am
              + tae[ie + 1, im + 1] * ae * am)
        Ta = Ma + s * Ta

        if (Ta < 0.0):
            Ta = Ta + TWO_PI

    z = a * (1.0 - e ** 2) / (1.0 + e * cos(Ta)) * sqrt(1.0 - sin(w + Ta) ** 2 * sin(i) ** 2)
    z *= copysign(1.0, sin(w + Ta))
    return z


@njit(cache=cache)
def z_newton_s(t, pv):
    """Normalized projected distance for scalar t.

    pv = [t0, p, a, i, e, w]
    """
    t0, p, a, i, e, w = pv
    Ta = ta_newton_s(t, t0, p, e, w)
    return z_from_ta_s(Ta, a, i, e, w)

@njit(cache=cache)
def z_newton_v(ts, pv):
    t0, p, a, i, e, w = pv
    Ta = ta_newton_v(ts, t0, p, e, w)
    return z_from_ta_v(Ta, a, i, e, w)

@njit(parallel=True, fastmath=True)
def z_newton_p(ts, pv):
    t0, p, a, i, e, w = pv
    zs = zeros_like(ts)
    for j in prange(len(ts)):
        t = ts[j]
        ma_offset = arctan2(sqrt(1.0-e**2) * sin(HALF_PI - w), e + cos(HALF_PI - w))
        ma_offset -= e*sin(ma_offset)
        Ma = mod(TWO_PI * (t - (t0 - ma_offset * p / TWO_PI)) / p, TWO_PI)
        Ea = Ma
        err = 0.05
        k = 0
        while abs(err) > 1e-8 and k<1000:
            err   = Ea - e*sin(Ea) - Ma
            Ea = Ea - err/(1.0-e*cos(Ea))
            k += 1
        sta = sqrt(1.0-e**2) * sin(Ea)/(1.0-e*cos(Ea))
        cta = (cos(Ea)-e)/(1.0-e*cos(Ea))
        Ta  = arctan2(sta, cta)
        z  = a*(1.0-e**2)/(1.0+e*cos(Ta)) * sqrt(1.0 - sin(w+Ta)**2 * sin(i)**2)
        z *= copysign(1.0, sin(w+Ta))
        zs[j] = z
    return zs

@njit(parallel=True, fastmath=True)
def z_newton_mp(ts, pvs):
    zs = zeros((pvs.shape[0], ts.size))
    for j in prange(zs.size):
        ipv, ipt = j // ts.size, j % ts.size
        t0, p, a, i, e, w = pvs[ipv]
        t = ts[ipt]
        ma_offset = arctan2(sqrt(1.0-e**2) * sin(HALF_PI - w), e + cos(HALF_PI - w))
        ma_offset -= e*sin(ma_offset)
        Ma = mod(TWO_PI * (t - (t0 - ma_offset * p / TWO_PI)) / p, TWO_PI)
        Ea = Ma
        err = 0.05
        k = 0
        while abs(err) > 1e-8 and k<1000:
            err   = Ea - e*sin(Ea) - Ma
            Ea = Ea - err/(1.0-e*cos(Ea))
            k += 1
        sta = sqrt(1.0-e**2) * sin(Ea)/(1.0-e*cos(Ea))
        cta = (cos(Ea)-e)/(1.0-e*cos(Ea))
        Ta  = arctan2(sta, cta)
        z  = a*(1.0-e**2)/(1.0+e*cos(Ta)) * sqrt(1.0 - sin(w+Ta)**2 * sin(i)**2)
        z *= copysign(1.0, sin(w+Ta))
        zs[ipv,ipt] = z
    return zs



# Z: Iteration
# ------------


@njit(cache=cache)
def z_iter_s(t, pv):
    t0, p, a, i, e, w = pv
    Ta = ta_iter_s(t, t0, p, e, w)
    return z_from_ta_s(Ta, a, i, e, w)


@njit(cache=cache)
def z_iter_v(ts, pv):
    t0, p, a, i, e, w = pv
    Ta = ta_iter_v(ts, t0, p, e, w)
    return z_from_ta_v(Ta, a, i, e, w)


@njit(parallel=True, fastmath=True)
def z_iter_p(ts, pv):
    t0, p, a, i, e, w = pv
    ma_offset = arctan2(sqrt(1.0 - e ** 2) * sin(HALF_PI - w), e + cos(HALF_PI - w))
    ma_offset -= e * sin(ma_offset)
    zs = zeros_like(ts)
    for j in prange(len(ts)):
        t = ts[j]
        Ma = mod(TWO_PI * (t - (t0 - ma_offset * p / TWO_PI)) / p, TWO_PI)
        ec = e*sin(Ma)/(1.0 - e*cos(Ma))
        for k in range(15):
            ect   = ec
            ec = e*sin(Ma+ec)
            if (abs(ect-ec) < 1e-4):
                break
        Ea  = Ma + ec
        sta = sqrt(1.0-e**2) * sin(Ea)/(1.0-e*cos(Ea))
        cta = (cos(Ea)-e)/(1.0-e*cos(Ea))
        Ta  = arctan2(sta, cta)
        z  = a*(1.0-e**2)/(1.0+e*cos(Ta)) * sqrt(1.0 - sin(w+Ta)**2 * sin(i)**2)
        z *= copysign(1.0, sin(w+Ta))
        zs[j] = z
    return zs

# Z: Series expansion
# -------------------


@njit(cache=cache)
def z_ps3(t, pv):
    t0, p, a, i, e, w = pv
    return z_from_ta_v(ta_ps3(t, t0, p, e, w), a, i, e, w)


@njit(cache=cache)
def z_ps5(t, pv):
    t0, p, a, i, e, w = pv
    return z_from_ta_v(ta_ps5(t, t0, p, e, w), a, i, e, w)


# Utility functions
# -----------------

@njit(cache=cache)
def impact_parameter(a, i):
    return a * cos(i)


@njit(cache=cache)
def impact_parameter_ec(a, i, e, w, tr_sign):
    return a * cos(i) * ((1.-e**2) / (1.+tr_sign*e*sin(w)))


@njit(cache=cache)
def duration_eccentric(p, k, a, i, e, w, tr_sign):
    b  = impact_parameter_ec(a, i, e, w, tr_sign)
    ae = sqrt(1.-e**2)/(1.+tr_sign*e*sin(w))
    return p/pi  * arcsin(sqrt((1.+k)**2-b**2)/(a*sin(i))) * ae


@njit(cache=cache)
def eclipse_phase(p, i, e, w):
    """ Phase for the secondary eclipse center.

    Exact secondary eclipse center phase, good for all eccentricities.
    """
    etr = arctan2(sqrt(1. - e**2) * sin(HALF_PI - w), e + cos(HALF_PI - w))
    eec = arctan2(sqrt(1. - e**2) * sin(HALF_PI + pi - w), e + cos(HALF_PI + pi - w))
    mtr = etr - e * sin(etr)
    mec = eec - e * sin(eec)
    phase = (mec - mtr) * p / TWO_PI
    return phase if phase > 0. else p + phase


@njit(cache=cache)
def eclipse_phase_ap(p, i, e, w):
    """Approximate phase for the secondary eclipse center.

    Approximate phase for the center of the secondary eclipse (ok for e < 0.5)
    (Kallrath, J., "Eclipsing Binary Stars: Modeling and Analysis", p. 85)
    """
    return 0.5 * p + (p * INV_PI * e * cos(w) * (1. + 1. / sin(i)**2))


@njit(cache=cache)
def p_from_am(a=1., ms=1.):
    """Orbital period from the semi-major axis and stellar mass.

    Parameters
    ----------

      a    : semi-major axis [AU]
      ms   : stellar mass    [M_Sun]

    Returns
    -------

      p    : Orbital period  [d]
    """
    return sqrt((4*pi**2*(a*au)**3)/(G*ms*msun)) / D_S


@njit(cache=cache)
def a_from_mp(ms, period):
    """Semi-major axis from the stellar mass and planet's orbital period.

    Parameters
    ----------

      ms     : stellar mass    [M_Sun]
      period : orbital period  [d]

    Returns
    -------

      a : semi-major axis [AU]
    """
    return ((G * (ms*msun) * (period * D_S)**2) / (4 * pi**2))**(1 / 3) / au


@njit(cache=cache)
def as_from_rhop(rho, period):
    """Scaled semi-major axis from the stellar density and planet's orbital period.

    Parameters
    ----------

      rho    : stellar density [g/cm^3]
      period : orbital period  [d]

    Returns
    -------

      as : scaled semi-major axis [R_star]
    """
    return (G/(3*pi))**(1/3) * ((period * D_S)**2 * 1e3 * rho)**(1 / 3)


@njit(cache=cache)
def rho_from_asp(a, period):
    """Stellar density from the scaled semi-major axis and orbital period.
    """
    return (3*pi*a**3)/(G * (period * D_S)**2) * 1e-3


@njit(cache=cache)
def a_from_rhoprs(rho, period, rstar):
    """Semi-major axis from the stellar density, stellar radius, and planet's orbital period.

    Parameters
    ----------

      rho    : stellar density [g/cm^3]
      period : orbital period  [d]
      rstar  : stellar radius  [R_Sun]

    Returns
    -------

      a : semi-major axis [AU]
    """
    return as_from_rhop(rho, period)*rstar*rsun/au


@njit(cache=cache)
def af_transit(e, w):
    """Calculates the -- factor during the transit"""
    return (1.0-e**2)/(1.0 + e*sin(w))


@njit(cache=cache)
def i_from_baew(b, a, e, w):
    """Orbital inclination from the impact parameter, scaled semi-major axis, eccentricity and argument of periastron

    Parameters
    ----------

      b  : impact parameter       [-]
      a  : scaled semi-major axis [R_Star]
      e  : eccentricity           [-]
      w  : argument of periastron [rad]

    Returns
    -------

      i  : inclination            [rad]
    """
    return arccos(b / (a*af_transit(e, w)))


@njit(cache=cache)
def i_from_ba(b, a):
    """Orbital inclination from the impact parameter and scaled semi-major axis.

    Parameters
    ----------

      b  : impact parameter       [-]
      a  : scaled semi-major axis [R_Star]

    Returns
    -------

      i  : inclination            [rad]
    """
    return arccos(b/a)


@njit(cache=cache)
def d_from_pkaiews(p, k, a, i, e, w, tr_sign, kind=14):
    """Transit duration (T14 or T23) from p, k, a, i, e, w, and the transit sign.

    Calculates the transit duration (T14) from the orbital period, planet-star radius ratio, scaled semi-major axis,
    orbital inclination, eccentricity, argument of periastron, and the sign of the transit (transit:1, eclipse: -1).

     Parameters
     ----------

       p  : orbital period         [d]
       k  : radius ratio           [R_Star]
       a  : scaled semi-major axis [R_star]
       i  : orbital inclination    [rad]
       e  : eccentricity           [-]
       w  : argument of periastron [rad]
       tr_sign : transit sign, 1 for a transit, -1 for an eclipse
       kind: either 14 for full transit duration or 23 for total transit duration

     Returns
     -------

       d  : transit duration T14  [d]
     """
    b  = impact_parameter_ec(a, i, e, w, tr_sign)
    ae = sqrt(1.-e**2)/(1.+tr_sign*e*sin(w))
    ds = 1. if kind == 14 else -1.
    return p/pi  * arcsin(sqrt((1.+ds*k)**2-b**2)/(a*sin(i))) * ae


@njit(cache=cache)
def p_from_dkaiews(d, k, a, i, e, w, tr_sign):
    """Orbital period from d, k, a, i, e, w, and the transit sign.

    Calculates the orbital period from the transit duration (T14), planet-star radius ratio, scaled semi-major axis,
    orbital inclination, eccentricity, argument of periastron, and the sign of the transit (transit:1, eclipse: -1).

     Parameters
     ----------

       d  : transit duration T14   [d]
       k  : radius ratio           [R_Star]
       a  : scaled semi-major axis [R_star]
       i  : orbital inclination    [rad]
       e  : eccentricity           [-]
       w  : argument of periastron [rad]
       tr_sign : transit sign, 1 for a transit, -1 for an eclipse


    Returns
    -------

      p  : orbital period         [d]
    """
    b  = impact_parameter_ec(a, i, e, w, tr_sign)
    ae = sqrt(1.-e**2)/(1.+tr_sign*e*sin(w))
    return (d*pi) / (arcsin(sqrt((1.+k)**2-b**2)/(a*sin(i))) * ae)


@njit
def as_from_dkp(d, p, k):
    """Assumes b=0"""
    return sqrt((1.0+k)**2) / sin(pi*d/p)


from numpy import sqrt, abs, zeros, ones
from numba import njit, prange

from .rrmodel import circle_circle_intersection_area
from .ldmodels import ld_quadratic, ldi_quadratic

@njit(parallel=True)
def quadratic_small_planet_zp(z, k, ldc):
    """Small planet approximation with quadratic limb darkening.

    Parallelised transit model using the small-planet approximation as a 
    function of normalized planet-star separation. The function is meant to
    be used to evaluate the model for npv radius ratios and limb darkening 
    coefficients in parallel, where npv should be relatively large to offset
    the cost from the threading initialization. 

    Parameters
    ----------
    z: array
        Planet-star separations as a 1d ndarray
    k: array
        Radius ratios as a 1d ndarray
    ldc: array
        Limb darkening coefficients as [npv, 2] ndarray

    Returns
    -------

    """
    npt = z.size
    npv = k.size

    bs = zeros(npt)
    mus = zeros(npt)
    for ipt in range(npt):
        bs[ipt] = abs(z[ipt])
        mus[ipt] = sqrt(1. - min(bs[ipt]**2, 1.0))

    flux = ones((npv, npt))
    for ipv in prange(npv):
        istar = ldi_quadratic(ldc[ipv])
        for ipt in range(npt):
            if bs[ipt] < 1.0 + k[ipv]:
                iplanet = ld_quadratic(mus[ipt], ldc[ipv])
                aplanet = circle_circle_intersection_area(1.0, k[ipv], bs[ipt])
                flux[ipv, ipt] = (istar - iplanet*aplanet) / istar
    return flux


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import scipy.ndimage as ndi
import pyopencl as cl
import seaborn as sb

from matplotlib.pyplot import subplots, setp
from numpy import sqrt, array, inf, int, s_, percentile, median, mean, round, zeros, isfinite, where, atleast_2d, ceil, \
    newaxis

from ..param.parameter import GParameter, LParameter
from ..param.parameter import UniformPrior as UP, NormalPrior as NP
from .oclttvlpf import OCLTTVLPF, plot_estimates
from ..orbits.orbits_py import as_from_rhop, i_from_ba, p_from_dkaiews

with_seaborn = True


class OCLTDVLPF(OCLTTVLPF):
    def __init__(self, target: str, zero_epoch: float, period: float, duration_prior: tuple,
                 passbands: list, times: list = None, fluxes: list = None, errors: list = None, pbids: list = None,
                 nsamples: int = 1, exptime: float = 0.020433598, cl_ctx=None, cl_queue=None):
        self.t14_prior = duration_prior
        super().__init__(target, zero_epoch, period, passbands, times, fluxes, errors, pbids, nsamples, exptime, cl_ctx,
                         cl_queue)

    def _init_p_orbit(self):
        """Orbit parameter initialisation for a TTV model.
           """

        # Basic orbital parameters
        # ------------------------
        porbit = [GParameter('rho', 'stellar_density', 'g/cm^3', UP(0.1, 25.0), (0, inf)),
                  GParameter('b', 'impact_parameter', 'R_s', UP(0.0, 1.0), (0, 1))]

        # Transit centers
        # ---------------

        def create_tc_prior(t, f, p=5):
            m = f > percentile(f, p)
            m = ~ndi.binary_erosion(m, iterations=6, border_value=1)
            return NP(t[m].mean(), 0.25 * t[m].ptp())

        self.tnumber = round((array([t.mean() for t in self.times]) - self.zero_epoch) / self.period).astype(int)
        for t, f, tn in zip(self.times, self.fluxes, self.tnumber):
            prior = create_tc_prior(t, f, self._tc_prior_percentile)
            porbit.append(GParameter(f'tc_{tn:d}', f'transit_centre_{tn:d}', 'd', prior, (-inf, inf)))

        # Transit durations
        # -----------------
        for tn in self.tnumber:
            porbit.append(GParameter(f't14_{tn:d}', f'duration_{tn:d}', 'd', NP(*self.t14_prior), (0, inf)))

        self.ps.add_global_block('orbit', porbit)
        self._start_tc = 2
        self._sl_tc = s_[self._start_tc:self._start_tc + self.nlc]
        self._start_td = 2 + self.nlc
        self._sl_td = s_[self._start_td:self._start_td + self.nlc]

    def transit_model(self, pvp, copy=False):
        pvp = atleast_2d(pvp)
        pvp_cl = zeros([pvp.shape[0], 5 + 2 * self.nlc], "f")  # k [tc] [p] a i e w
        uv = zeros([pvp.shape[0], 2], "f")
        tc_end = 1 + self.nlc
        td_end = 1 + 2 * self.nlc
        pvp_cl[:, 0:1] = k = sqrt(pvp[:, self._pid_k2])  # Radius ratio
        pvp_cl[:, 1:tc_end] = pvp[:, self._sl_tc]  # Transit centres
        pvp_cl[:, td_end + 0] = a = as_from_rhop(pvp[:, 0], self.period)
        pvp_cl[:, td_end + 1] = i = i_from_ba(pvp[:, 1], a)
        pvp_cl[:, tc_end:td_end] = p_from_dkaiews(pvp[:, self._sl_td], k, a[:, newaxis], i[:, newaxis], 0., 0.,
                                                  1)  # Orbital periods
        a, b = sqrt(pvp[:, self._sl_ld][:, 0]), 2. * pvp[:, self._sl_ld][:, 1]
        uv[:, 0] = a * b
        uv[:, 1] = a * (1. - b)
        flux = self.tm.evaluate_t_pv2d_ttv(self.timea, pvp_cl, uv, self.lcids, self.nlc, copy=copy, tdv=True)
        return flux.T if copy else None


    def plot_tdvs(self, burn=0, thin=1, axs=None, figsize=None, bwidth=0.8, fmt='h', windows=None):
        assert fmt in ('d', 'h', 'min')
        multiplier = {'d': 1, 'h': 24, 'min': 1440}
        ncol = 1 if windows is None else len(windows)
        fig, axs = (None, axs) if axs is not None else subplots(1, ncol, figsize=figsize, sharey=True)
        df = self.posterior_samples(burn, thin)
        dcols = [c for c in df.columns if 't14_' in c]
        p = multiplier[fmt] * df[dcols].quantile([0.50, 0.16, 0.84, 0.005, 0.995]).values
        setp(axs, ylabel='Transit duration [{}]'.format(fmt), xlabel='Transit number')

        if windows is None:
            plot_estimates(self.tnumber, p, axs)
        else:
            setp(axs[1:], ylabel='')
            for ax, w in zip(axs, windows):
                m = (self.tnumber > w[0]) & (self.tnumber < w[1])
                plot_estimates(self.tnumber[m], p[:, m], ax)
                setp(ax, xlim=w)
                if with_seaborn:
                    sb.despine(ax=ax, offset=15)

        if fig:
            fig.tight_layout()

        return axs



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pandas as pd

from itertools import product
from numpy import inf, array, zeros, unique, where, stack, atleast_2d, squeeze, all

from .prior import DefaultPrior, NormalPrior, UniformPrior, JeffreysPrior, LogLogisticPrior, GammaPrior


class Parameter:
    def __init__(self, name, description='', unit='', prior=None, bounds=(-inf, inf), **kwargs):
        self.name = name
        self.description = description
        self.unit = unit
        self.prior = prior or DefaultPrior()
        self.bounds = bounds
        self.pid = -1
        self.scope = kwargs.get('scope', 'global')
        assert self.scope in ['global', 'local', 'passband']

    def __str__(self):
        return f"{self.pid:3d} |{self.scope[0].upper():1s}| {self.name:14s} {str(self.prior):40} [{self.bounds[0]:8.2f} .. {self.bounds[1]:8.2f}]"

    def __repr__(self):
        return str(self)

    def truncated_lnprior(self, v):
        if self.bounds[0] < v < self.bounds[1]:
            return self.prior.logpdf(v)
        else:
            return -inf

    def lnprior(self, v):
        return self.prior.logpdf(v)

    def rvs(self, size):
        return self.prior.rvs(size)


class GParameter(Parameter):
    def __init__(self, name, description='', unit='', prior=None, bounds=(-inf, inf), **kwargs):
        super().__init__(name, description, unit, prior, bounds, scope='global', **kwargs)


class LParameter(Parameter):
    def __init__(self, name, description='', unit='', prior=None, bounds=(-inf, inf), **kwargs):
        super().__init__(name, description, unit, prior, bounds, scope='local', **kwargs)


class PParameter(Parameter):
    def __init__(self, name, description='', unit='', prior=None, bounds=(-inf, inf), **kwargs):
        super().__init__(name, description, unit, prior, bounds, scope='passband', **kwargs)


class GParameterBlock:
    def __init__(self, name, start, stop):
        self.name = name
        self.start = start
        self.stop = stop
        self.slice = slice(self.start, self.stop)

    def __repr__(self):
        return "GParameterBlock('{}', {}, {})".format(self.name, self.start, self.stop)


class PParameterBlock(GParameterBlock):
    def __init__(self, name, start, stop, isize, npb):
        assert (stop-start) // isize == npb, "The number of parameter sets != npb"
        assert (stop-start) % isize == 0
        super(PParameterBlock, self).__init__(name, start, stop)
        self.isize = isize
        self.nelem = nelem = (stop-start) // isize
        self.slice = slice(self.start, self.stop)
        self.slices = [slice(self.start+isize*i, self.start+isize*(i+1)) for i in range(nelem)]

    def __repr__(self):
        return "<PParameterBlock('{}', {}, {}, {})".format(self.name, self.start, self.stop, self.isize)


class LParameterBlock(GParameterBlock):
    def __init__(self, name, start, stop, isize, nlc):
        assert (stop-start) // isize == nlc, "The number of parameter sets != nlc "
        assert (stop-start) % isize == 0
        super(LParameterBlock, self).__init__(name, start, stop)
        self.isize = isize
        self.nelem = nelem = (stop-start) // isize
        self.slice = slice(self.start, self.stop)
        self.slices = [slice(self.start+isize*i, self.start+isize*(i+1)) for i in range(nelem)]

    def __repr__(self):
        return "<LParameterBlock('{}', {}, {}, {})".format(self.name, self.start, self.stop, self.isize)


class ParameterSet(list):
    def __init__(self, *args):
        super().__init__(*args)
        self.blocks = []
        self.bounds = None
        self.frozen = False

    def add_global_block(self, name, pars):
        start = len(self)
        stop = start + len(pars)
        self.blocks.append(GParameterBlock(name, start, stop))
        self.extend(pars)

    def add_passband_block(self, name, isize, npb, pars):
        start = len(self)
        stop = start + len(pars)
        self.blocks.append(PParameterBlock(name, start, stop, isize, npb))
        self.extend(pars)

    def add_lightcurve_block(self, name, isize, nlc, pars):
        start = len(self)
        stop = start + len(pars)
        self.blocks.append(LParameterBlock(name, start, stop, isize, nlc))
        self.extend(pars)

    def append(self, *args):
        if not self.frozen:
            super().append(*args)
        else:
            raise ValueError('Trying to append to a frozen ParameterSet')

    def extend(self, *args):
        if not self.frozen:
            super().extend(*args)
        else:
            raise ValueError('Trying to extend a frozen ParameterSet')

    def _update_indices(self):
        if not self.frozen:
            self.bounds = array([p.bounds for p in self])
            self.lbounds = self.bounds[:, 0]
            self.ubounds = self.bounds[:, 1]
            for i, p in enumerate(self):
                p.pid = i
        else:
            raise ValueError('Trying to update a frozen ParameterSet')

    def lnprior(self, pv):
        pv = atleast_2d(pv)
        lnp = zeros(pv.shape[0])
        m = all(pv > self.bounds[:, 0], 1) & all(pv < self.bounds[:, 1], 1)
        for i, p in enumerate(self):
            lnp += p.lnprior(pv[:, i])
        return squeeze(where(m, lnp, -inf))

    def freeze(self):
        self._update_indices()
        self.frozen = True

    def thaw(self):
        self.frozen = False

    def validate(self):
        # General tests
        # -------------
        if not all(pd.Categorical(self.names).value_counts() == 1):
            raise ValueError('Duplicate parameter names are not allowed')

        # Validate global parameters
        # --------------------------

        # Validate the passband-dependent parameters
        # ------------------------------------------
        pnames = self.passband_parameters.names

        if not all(['_' in name for name in pnames]):
            raise ValueError('Passband-dependent parameters must be named as parameter_passband')

        pars = unique(list(n.split('_')[0] for n in pnames))
        pbs = unique(list(n.split('_')[1] for n in pnames))

        if not len(self.passband_parameters) == len(pars) * len(pbs):
            raise ValueError('Passband-dependent parameters must exist for all passbands')

        if not all(['{}_{}'.format(pr, pb) in pnames for pr, pb in product(pars, pbs)]):
            raise ValueError('Passband-dependent parameters must be named as parameter_passband')

        # Validate the per-light-curve parameters
        # ---------------------------------------


    def find_pid(self, name):
        for p in self:
            if name == p.name:
                return p.pid
        raise KeyError('Could not find parameter {}'.format(name))


    def sample_from_prior(self, size=1):
        return stack([p.rvs(size) for p in self.priors], 1)

    def check_pv(self, pv):
        for i, p in enumerate(self):
            lnp = p.prior.logpdf(pv[i])
            b = self.bounds[i]
            is_finite = p.bounds[0] < pv[i] < p.bounds[1]
            print(f"|{p.pid:3d}| {'*' if not is_finite else ' '} {p.name:10} {b[0]:7.1f} < {pv[i]:10.2f} < {b[1]:<8.1f} {str(p.prior):40} log P = {lnp:6.2f}")

    @property
    def mean_pv(self):
        x0 = zeros(len(self))
        for i, p in enumerate(self.priors):
            if isinstance(p, NormalPrior):
                x0[i] = p.mean
            elif isinstance(p, UniformPrior):
                x0[i] = 0.5 * (p.a + p.b)
            else:
                raise ValueError
        return x0

    @property
    def names(self):
        return [p.name for p in self]

    @property
    def units(self):
        return [p.unit for p in self]

    @property
    def descriptions(self):
        return [p.description for p in self]

    @property
    def priors(self):
        return [p.prior for p in self]

    @property
    def scopes(self):
        return [p.scope for p in self]

    @property
    def global_parameters(self):
        return ParameterSet([p for p in self if p.scope == 'global'])

    @property
    def local_parameters(self):
        return ParameterSet([p for p in self if p.scope == 'local'])

    @property
    def passband_parameters(self):
        return ParameterSet([p for p in self if p.scope == 'passband'])


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Astrophysical quantities

 Astrophysical quantities with their error estimates
 ===================================================

 Notes
 -----
  Error estimates follow the form Q_e where Q is the
  quantity in question, and are available only for a
  subset of quantities.

 =============  ============================== ======
 ``au``         astronomical unit                   m
 ``msun``       solar mass                         kg
 ``rsun``       solar radius                        m
 ``dsun``       mean solar mass density        g/cm^3
 ``mjup``       Jupiter mass                       kg
 ``rjup``       volumetric mean Jupiter radius      m
 ``rjup_eq``    equatorial Jupiter radius           m
 ``djup``       mean Jupiter mass density      g/cm^3
 ``mnep``       Neptune mass            radius     kg
 ``rnep``       volumetric mean Neptune radius      m
 ``rnep_eq``    equatorial Neptune radius           m
 ``djup``       mean Neptune mass density      g/cm^3
 ``rearth``     volumetric mean Earth radius        m
 ``rearth_eq``  equatorial Earth radius             m
 =============  ============================== ======
"""

from __future__ import division

d_h = 24.
d_m = 60 * d_h
d_s = 60 * d_m

au,   au_e             = 1.496e11, 0.0       
msun, msun_e           = 1.9891e30, 0.0      
rsun, rsun_e           = 0.5*1.392684e9, 0.0
dsun, dsun_e           = 1.408, 0.0
mjup, mjup_e           = 1.89896e27, 0.0     
rjup, rjup_e           = 6.9911e7, 0.0       
rjup_eq, rjup_eq_e     = 7.1492e7, 0.0
djup, djup_e           = 1.326, 0.0       
mnep, mnep_e           = 1.0242e26, 0.0
rnep, mnep_e           = 2.4622e7, 0.0
rnep_eq, mnep_eq_e     = 2.4764e7, 0.0
dnep, dnep_e           = 1.638, 0.0       
mearth, mearth_e       = 5.9726e24, 0.0
rearth, rearth_e       = 6.371e6, 0.0
rearth_eq, rearth_eq_e = 6.3781e6, 0.0
dearth, dearth_e       = 5.514, 0.0       



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path

from astropy.table import Table
from numba import njit, prange
from numpy import atleast_2d, zeros, log, concatenate, pi, transpose, sum, compress, ones
from uncertainties import UFloat, ufloat

from .oclttvlpf import OCLTTVLPF
from ..utils.keplerlc import KeplerLC
from .baselines.legendrebaseline import  LegendreBaseline

@njit(parallel=True, cache=False, fastmath=True)
def lnlike_normal_v(o, m, e):
    m = atleast_2d(m)
    npv = m.shape[0]
    npt = o.size
    lnl = zeros(npv)
    for i in prange(npv):
        lnl[i] = -npt*log(e[i, 0]) - 0.5*log(2*pi) - 0.5*sum(((o-m[i, :])/e[i ,0])**2)
    return lnl


class TESSCLTTVLPF(LegendreBaseline, OCLTTVLPF):
    def __init__(self, name: str, dfile: Path, zero_epoch: float, period: float, nsamples: int = 10,
                 trdur: float = 0.125, bldur: float = 0.3, nlegendre: int = 2, ctx = None, queue = None):

        zero_epoch = zero_epoch if isinstance(zero_epoch, UFloat) else ufloat(zero_epoch, 1e-5)
        period = period if isinstance(period, UFloat) else ufloat(period, 1e-7)

        tb = Table.read(dfile)
        self.bjdrefi = tb.meta['BJDREFI']
        zero_epoch = zero_epoch - self.bjdrefi

        df = tb.to_pandas().dropna(subset=['TIME', 'SAP_FLUX', 'PDCSAP_FLUX'])
        self.lc = lc = KeplerLC(df.TIME.values, df.SAP_FLUX.values, zeros(df.shape[0]),
                                zero_epoch.n, period.n, trdur, bldur)

        LegendreBaseline.__init__(self, nlegendre)
        OCLTTVLPF.__init__(self, name, zero_epoch, period, ['TESS'],
                         times=lc.time_per_transit, fluxes=lc.normalized_flux_per_transit,
                         pbids=zeros(lc.nt, 'int'), nsamples=nsamples, exptimes=[0.00139],
                         cl_ctx=ctx, cl_queue=queue)

        self.lnlikelihood = self.lnlikelihood_nb

    def create_pv_population(self, npop=50):
        pvp = self.ps.sample_from_prior(npop)
        return pvp

    def flux_model(self, pvp):
        tmodel = transpose(self.transit_model(pvp, copy=True)).copy()
        return tmodel * self.baseline(pvp)

    def lnlikelihood_nb(self, pvp):
        fmodel = self.flux_model(pvp).astype('d')
        err = 10**atleast_2d(pvp)[:, self._sl_err]
        return lnlike_normal_v(self.ofluxa, fmodel, err)



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""PyTransit: fast and easy exoplanet transit modelling in Python

This package offers Python interfaces for a set of exoplanet transit light curve
models implemented in Python (with Numba acceleration) and OpenCL.

Author
  Hannu Parviainen  <hannu@iac.es>

Date
  9.08.2020

"""

from .version import __version__

# Generic
# -------
from .models.transitmodel import TransitModel
from .contamination.filter import DeltaFilter, BoxcarFilter, TabulatedFilter, sdss_g, sdss_r, sdss_i, sdss_z

# Numba models
# ------------
from .models.qpower2 import QPower2Model
from .models.ma_quadratic import QuadraticModel
from .models.ma_uniform import UniformModel
from .models.eclipse_model import EclipseModel
from .models.ma_chromosphere import ChromosphereModel
from .models.general import GeneralModel
from .models.rrmodel import RoadRunnerModel
from .models.swiftmodel import SwiftModel
from .models.osmodel import OblateStarModel
from .models.gdmodel import GravityDarkenedModel

# OpenCL models
# -------------
class DummyModelCL:
    def __init__(self, *args, **kwargs):
        raise ImportError('Cannot use the OpenCL models because pyopencl is not installed. Please install pyopencl.')


try:
    from .models.qpower2_cl import QPower2ModelCL
    from .models.ma_quadratic_cl import QuadraticModelCL
    from .models.ma_uniform_cl import UniformModelCL
    from .models.swiftmodel_cl import SwiftModelCL, SwiftModelCL as SWIFTModelCL
except ModuleNotFoundError:
    QPower2ModelCL = DummyModelCL
    QuadraticModelCL = DummyModelCL
    UniformModelCL = DummyModelCL
    SwiftModelCL = SWIFTModelCL = DummyModelCL


# LDTk limb darkening for the Swift model
# ---------------------------------------
class DummyLDTkLDModel:
    def __init__(self, *args, **kwargs):
        raise ImportError('Cannot use the LDTk limb darkening model because ldtk is not installed. Please install ldtk.')


try:
    from .models.ldtkldm import LDTkLDModel, LDTkLD
except ModuleNotFoundError:
    LDTkLD = LDTkLDModel = DummyLDTkLDModel


# Log posterior functions
# -----------------------
from .lpf.lpf import BaseLPF
from .lpf.transitlpf import TransitLPF
from .lpf.cntlpf import PhysContLPF
from .lpf.baselines.legendrebaseline import LegendreBaseline
from .lpf.baselines.linearbaseline import LinearModelBaseline


# Utilities
# ---------
from .param.parameter import UniformPrior, NormalPrior
from .utils import md_rs_from_rho



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Iterable, Optional

from astropy.table import Column
from corner import corner
from matplotlib.pyplot import subplots, setp
from numba import njit
from numpy import zeros, log, pi, inf, atleast_2d, arange, arctan2, cos, squeeze, median, linspace, \
    percentile, argsort, sum, concatenate, full, sqrt, ndarray
from numpy.random.mtrand import permutation

from pytransit.lpf.logposteriorfunction import LogPosteriorFunction
from pytransit.utils.misc import fold
from pytransit.param import GParameter, UniformPrior as UP, NormalPrior as NP, ParameterSet
from pytransit.orbits.orbits_py import ta_newton_v


@njit(cache=False)
def lnlike_normal_s(o, m, e):
    return -sum(log(e)) - 0.5*o.size*log(2.*pi) - 0.5*sum((o-m)**2/e**2)


@njit(cache=False)
def lnlike_normal(o, m, e):
    npv = m.shape[0]
    lnl = zeros(npv)
    for ipv in range(npv):
        lnl[ipv] = -sum(log(e[ipv])) - 0.5*o.size*log(2.*pi) - 0.5*sum((o-m[ipv])**2/e[ipv]**2)
    return lnl


class RVLPF(LogPosteriorFunction):
    def __init__(self, name: str, nplanets: int, times, rvs, rves, rvis=None):
        super().__init__(name)

        def transform_input(a):
            if isinstance(a, (list, tuple)):
                return a
            elif isinstance(a, ndarray):
                return [a]
            elif isinstance(a, Column):
                return [a.data]

        times = transform_input(times)
        rvs = transform_input(rvs)
        rves = transform_input(rves)
        if rvis:
            rvis = transform_input(rvis)
        else:
            rvis = zeros(len(times), 'int')

        self._tref = concatenate(times).mean()
        self.rvm: RVModel = RVModel(self, nplanets, times, rvs, rves, rvis)

    def model(self, pv):
        return self.rvm.rv_model(pv)

    def lnlikelihood(self, pv):
        return squeeze(self.rvm.lnlikelihood(pv))

    def plot_rv_vs_time(self, method='de', pv=None, nsamples: int = 200, ntimes: int = 500, axs=None):
        return self.rvm.plot_rv_vs_time(method, pv, nsamples, ntimes, axs)

    def plot_rv_vs_phase(self, planet: int, method='de', pv=None, nsamples: int = 200, ntimes: int = 500, axs=None):
        return self.rvm.plot_rv_vs_phase(planet, method, pv, nsamples, ntimes, axs)

    def plot_posteriors(self):
        df = self.posterior_samples()
        labels = [f"{d} [{u}]" if u else d for d,u in zip(self.ps.descriptions, self.ps.units)]
        return corner(df, labels=labels)


class RVModel:
    """Radial velocity model.
    """

    def __init__(self, lpf: LogPosteriorFunction, nplanets: int,
                 times, rvs: Iterable, rves: Iterable, rvis: Iterable,
                 tref: Optional[float] = None):
        self.lpf = lpf

        if hasattr(lpf, 'nplanets'):
            assert lpf.nplanets == nplanets
        self.nplanets = nplanets

        if hasattr(lpf, '_tref'):
            assert tref is None
            self._tref = lpf._tref
        else:
            self._tref = lpf._tref = tref if tref is not None else 0.0

        self.times = None  # Mid-measurement times
        self.rvs = None  # RV values
        self.rves = None  # RV errors
        self.rvis = None  # RV instrument indices

        self._timea = None
        self._rva = None
        self._rvea = None
        self._rv_ids = None

        self.setup_data(times, rvs, rves, rvis)

        if hasattr(lpf, 'ps') and lpf.ps is not None:
            self.ps = lpf.ps
        else:
            self._init_parameters()
            lpf.ps = self.ps

        self._init_p_rv()

    def setup_data(self, times: Iterable, rvs: Iterable, rves: Iterable, rvis: Iterable):
        self.times = times
        self.rvs = rvs
        self.rves = rves
        self.rvis = rvis

        self._timea = concatenate(times) - self._tref
        self._rva = concatenate(rvs)
        self._rvea = concatenate(rves)
        self._rv_ids = concatenate([full(c.size, i) for i, c in enumerate(self.times)])

    def _init_parameters(self):
        self.ps = ps = ParameterSet([])
        pp = []
        for i in range(1, self.nplanets + 1):
            pp.extend([
                GParameter(f'tc_{i}', f'zero epoch {i}', 'd', NP(0.0, 0.1), (-inf, inf)),
                GParameter(f'p_{i}', f'period {i}', 'd', NP(1.0, 1e-5), (0, inf)),
                GParameter(f'secw_{i}', f'sqrt(e) cos(w) {i}', '', UP(-1.0, 1.0), (-1, 1)),
                GParameter(f'sesw_{i}', f'sqrt(e) sin(w) {i}', '', UP(-1.0, 1.0), (-1, 1)),
            ])
        ps.add_global_block('planets', pp)
        self._start_pl = ps.blocks[-1].start
        self._sl_pl = ps.blocks[-1].slice

    def _init_p_rv(self):
        self.ps.thaw()
        ps = self.ps
        prv = [GParameter(f'rv_shift_{self.rvis[i]}', f'systemic velocity {self.rvis[i]}', 'm/s', NP(0.0, 0.1), (-inf, inf)) for i in
               range(len(self.times))]
        ps.add_global_block('rv_shifts', prv)
        self._start_rvs = ps.blocks[-1].start
        self._sl_rvs = ps.blocks[-1].slice

        prv = [GParameter(f'rv_err_{self.rvis[i]}', 'additional rv error', 'm/s', UP(0.0, 1.0), (-inf, inf)) for i in
               range(len(self.times))]
        ps.add_global_block('rv_errors', prv)
        self._start_rv_err = ps.blocks[-1].start
        self._sl_rv_err = ps.blocks[-1].slice

        prv = []
        for i in range(1, self.nplanets + 1):
            prv.append(GParameter(f'rv_k_{i}', f'rv semiamplitude {i}', 'm/s', UP(0.0, 1.0), (0, inf)))
        ps.add_global_block('rv_semiamplitudes', prv)
        self._start_rvk = ps.blocks[-1].start
        self._sl_rvk = ps.blocks[-1].slice

        psl = [GParameter('rv_slope', 'linear rv slope', 'm/s', NP(0.0, 1.0), (-inf, inf))]
        ps.add_global_block('rv_slope', psl)
        self._start_rv_slope = ps.blocks[-1].start
        self._sl_rv_slope = ps.blocks[-1].slice
        self.ps.freeze()

        pnames = "rv_k_{} tc_{} p_{} secw_{} sesw_{}".split()
        self.pids = zeros((self.nplanets, 5), 'int')
        for ipl in range(self.nplanets):
            for ip, p in enumerate(pnames):
                name = p.format(ipl + 1)
                self.pids[ipl, ip] = self.ps.names.index(name)

    def rv_shifts(self, pvp):
        pvp = atleast_2d(pvp)
        return squeeze(pvp[:, self._sl_rvs][:, self._rv_ids])

    def rv_slope(self, pvp, times):
        pvp = atleast_2d(pvp)
        return times * pvp[:, self._sl_rv_slope]

    def rv_model(self, pvp, times=None, planets=None, add_sv=True, add_slope=True):
        times = self._timea if times is None else times - self._tref
        pvp = atleast_2d(pvp)
        rvs = zeros((pvp.shape[0], times.size))

        planets = planets if planets is not None else arange(self.nplanets)
        for ipl in planets:
            pv = pvp[:, self.pids[ipl]]
            tc = pv[:, 1] - self._tref
            p = pv[:, 2]
            e = pv[:, 3] ** 2 + pv[:, 4] ** 2
            w = arctan2(pv[:, 4], pv[:, 3])
            for ipv in range(pv.shape[0]):
                ta = ta_newton_v(times, t0=tc[ipv], p=p[ipv], e=e[ipv], w=w[ipv])
                rvs[ipv] += pv[ipv, 0] * (cos(w[ipv] + ta) + e[ipv] * cos(w[ipv]))
        if add_sv:
            rvs += self.rv_shifts(pvp)
        if add_slope:
            rvs += self.rv_slope(pvp, times)
        return squeeze(rvs)

    def lnlikelihood(self, pvp):
        if pvp.ndim == 2:
            errors = sqrt(self._rvea**2 + pvp[:, self._sl_rv_err][:, self._rv_ids]**2)
            return lnlike_normal(self._rva, self.rv_model(pvp), errors)
        else:
            errors = sqrt(self._rvea**2 + pvp[self._sl_rv_err][self._rv_ids]**2)
            return lnlike_normal_s(self._rva, self.rv_model(pvp), errors)

    def plot_rv_vs_time(self, method='de', pv=None, nsamples: int = 200, ntimes: int = 500, axs=None):

        if axs is None:
            fig, axs = subplots(2, 1, gridspec_kw={'height_ratios': (3, 1)}, sharex='all')
        else:
            fig, axs = None, axs

        if pv is None:
            if method == 'de':
                if self.lpf.de is None:
                    raise ValueError("The global optimizer hasn't been initialized.")
                pvp = None
                pv = self.lpf.de.minimum_location
            elif method == 'mcmc':
                if self.lpf.sampler is None:
                    raise ValueError("The sampler hasn't been initialized.")
                df = self.lpf.posterior_samples()
                pvp = permutation(df.values)[:nsamples, :]
                pv = median(pvp, 0)
        else:
            if pv.ndim == 1:
                pvp = None
                pv = pv
            else:
                pvp = permutation(pv)[:nsamples, :]
                pv = median(pvp, 0)

        rv_time = linspace(self._timea.min() - 1, self._timea.max() + 1, num=ntimes) + self._tref

        if pvp is None:
            rv_model = self.rv_model(pv, rv_time, add_sv=False)
            rv_model_limits = None
        else:
            rv_percentiles = percentile(self.rv_model(pvp, rv_time, add_sv=False), [50, 16, 84, 2.5, 97.5], 0)
            rv_model = rv_percentiles[0]
            rv_model_limits = rv_percentiles[1:]

        if rv_model_limits is not None:
            axs[0].fill_between(rv_time, rv_model_limits[2], rv_model_limits[3], facecolor='blue', alpha=0.25)
            axs[0].fill_between(rv_time, rv_model_limits[0], rv_model_limits[1], facecolor='darkblue', alpha=0.5)

        axs[0].plot(rv_time, rv_model, 'k', lw=1)
        axs[0].errorbar(self._timea + self._tref, self._rva + self.rv_shifts(pv), self._rvea, fmt='ok')
        axs[1].errorbar(self._timea + self._tref, self._rva - self.rv_model(pv), self._rvea, fmt='ok')

        if fig is not None:
            fig.tight_layout()
        return fig

    def plot_rv_vs_phase(self, planet: int, method='de', pv=None, nsamples: int = 200, ntimes: int = 500, axs=None):
        if axs is None:
            fig, axs = subplots(2, 1, gridspec_kw={'height_ratios': (3, 1)}, sharex='all')
        else:
            fig, axs = None, axs

        if pv is None:
            if method == 'de':
                if self.lpf.de is None:
                    raise ValueError("The global optimizer hasn't been initialized.")
                pvp = None
                pv = self.lpf.de.minimum_location
            elif method == 'mcmc':
                if self.lpf.sampler is None:
                    raise ValueError("The sampler hasn't been initialized.")
                df = self.lpf.posterior_samples()
                pvp = permutation(df.values)[:nsamples, :]
                pv = median(pvp, 0)
        else:
            if pv.ndim == 1:
                pvp = None
                pv = pv
            else:
                pvp = permutation(pv)[:nsamples, :]
                pv = median(pvp, 0)

        rv_time = linspace(self._timea.min() - 1, self._timea.max() + 1, num=ntimes)

        all_planets = set(range(self.nplanets))
        other_planets = all_planets.difference([planet])

        if pvp is None:
            rv_model = self.rv_model(pv, rv_time + self._tref, [planet], add_sv=False)
            rv_others = self.rv_model(pv, planets=other_planets, add_sv=False)
            rv_model_limits = None
        else:
            rv_percentiles = percentile(self.rv_model(pvp, rv_time + self._tref, [planet], add_sv=False),
                                        [50, 16, 84, 2.5, 97.5], 0)
            rv_model = rv_percentiles[0]
            rv_model_limits = rv_percentiles[1:]
            rv_others = median(self.rv_model(pvp, planets=other_planets, add_sv=False), 0)

        period = pv[self.ps.names.index(f'p_{planet + 1}')]
        tc = pv[self.ps.names.index(f'tc_{planet + 1}')] - self._tref

        phase = (fold(self._timea, period, tc, 0.5) - 0.5) * period
        phase_model = (fold(rv_time, period, tc, 0.5) - 0.5) * period
        msids = argsort(phase_model)

        if pvp is not None:
            axs[0].fill_between(phase_model[msids], rv_model_limits[2, msids], rv_model_limits[3, msids],
                                facecolor='blue',
                                alpha=0.15)
            axs[0].fill_between(phase_model[msids], rv_model_limits[0, msids], rv_model_limits[1, msids],
                                facecolor='darkblue',
                                alpha=0.25)

        axs[0].errorbar(phase, self._rva - rv_others - self.rv_shifts(pv), self._rvea, fmt='ok')
        axs[0].plot(phase_model[msids], rv_model[msids], 'k')
        axs[1].errorbar(phase, self._rva - self.rv_model(pv), self._rvea, fmt='ok')

        setp(axs[0], ylabel='RV [m/s]')
        setp(axs[1], xlabel='Phase [d]', ylabel='O-M [m/s]')

        axs[0].autoscale(axis='x', tight=True)
        if fig is not None:
            fig.tight_layout()
        return fig



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2021  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
from typing import List, Union, Optional
from astropy.table import Table
import astropy.io.fits as pf
from numpy import concatenate, diff, sqrt, full, median, array


def read_tess_spoc(tic: int,
                   datadir: Union[Path, str],
                   sectors: Optional[Union[List[int], str]] = 'all',
                   use_pdc: bool = False,
                   remove_contamination: bool = True):

    def file_filter(f, partial_tic, sectors):
        _, sector, tic, _, _ = f.name.split('-')
        if sectors != 'all':
            return int(sector[1:]) in sectors and str(partial_tic) in tic
        else:
            return str(partial_tic) in tic

    files = [f for f in sorted(Path(datadir).glob('tess*_lc.fits')) if file_filter(f, tic, sectors)]
    fcol = 'PDCSAP_FLUX' if use_pdc else 'SAP_FLUX'
    times, fluxes, sectors = [], [], []
    for f in files:
        tb = Table.read(f)
        bjdrefi = tb.meta['BJDREFI']
        df = tb.to_pandas().dropna(subset=['TIME', 'SAP_FLUX', 'PDCSAP_FLUX'])
        times.append(df['TIME'].values.copy() + bjdrefi)
        fluxes.append(array(df[fcol].values, 'd'))
        fluxes[-1] /= median(fluxes[-1])
        if use_pdc and not remove_contamination:
            contamination = 1 - tb.meta['CROWDSAP']
            fluxes[-1] = contamination + (1 - contamination) * fluxes[-1]
        sectors.append(full(fluxes[-1].size, pf.getval(f, 'sector')))

    return (concatenate(times), concatenate(fluxes), concatenate(sectors),
            [diff(f).std() / sqrt(2) for f in fluxes])



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Implements the differential evolution optimization method by Storn & Price
(Storn, R., Price, K., Journal of Global Optimization 11: 341--359, 1997)

.. moduleauthor:: Hannu Parviainen <hpparvi@gmail.com>
"""

from numba import njit
from numpy import asarray, zeros, zeros_like, tile, array, argmin, mod
from numpy.random import random, randint, rand, seed as rseed, uniform


def wrap(v, vmin, vmax):
    w = vmax - vmin
    return vmin + mod(asarray(v) - vmin, w)


@njit
def evolve_vector(i, pop, f, c):
    npop, ndim = pop.shape

    # --- Vector selection ---
    v1, v2, v3 = i, i, i
    while v1 == i:
        v1 = randint(npop)
    while (v2 == i) or (v2 == v1):
        v2 = randint(npop)
    while (v3 == i) or (v3 == v2) or (v3 == v1):
        v3 = randint(npop)

    # --- Mutation ---
    v = pop[v1] + f * (pop[v2] - pop[v3])

    # --- Cross over ---
    jf = randint(ndim)
    co = rand(ndim)
    for j in range(ndim):
        if co[j] > c and j != jf:
            v[j] = pop[i, j]
    return v


@njit
def evolve_population(pop, pop2, f, c):
    npop, ndim = pop.shape

    for i in range(npop):

        # --- Vector selection ---
        v1, v2, v3 = i, i, i
        while v1 == i:
            v1 = randint(npop)
        while (v2 == i) or (v2 == v1):
            v2 = randint(npop)
        while (v3 == i) or (v3 == v2) or (v3 == v1):
            v3 = randint(npop)

        # --- Mutation ---
        v = pop[v1] + f * (pop[v2] - pop[v3])

        # --- Cross over ---
        co = rand(ndim)
        for j in range(ndim):
            if co[j] <= c:
                pop2[i, j] = v[j]
            else:
                pop2[i, j] = pop[i, j]

        # --- Forced crossing ---
        j = randint(ndim)
        pop2[i, j] = v[j]

    return pop2


class DiffEvol(object):
    """
    Implements the differential evolution optimization method by Storn & Price
    (Storn, R., Price, K., Journal of Global Optimization 11: 341--359, 1997)

    :param fun:
       the function to be minimized

    :param bounds:
        parameter bounds as [npar,2] array

    :param npop:
        the size of the population (5*D - 10*D)

    :param  f: (optional)
        the difference amplification factor. Values of 0.5-0.8 are good in most cases.

    :param c: (optional)
        The cross-over probability. Use 0.9 to test for fast convergence, and smaller
        values (~0.1) for a more elaborate search.

    :param seed: (optional)
        Random seed

    :param maximize: (optional)
        Switch setting whether to maximize or minimize the function. Defaults to minimization.
    """

    def __init__(self, fun, bounds, npop, f=None, c=None, seed=None, maximize=False, vectorize=False, cbounds=(0.25, 1),
                 fbounds=(0.25, 0.75), pool=None, min_ptp=1e-2, args=[], kwargs={}):
        if seed is not None:
            rseed(seed)

        self.minfun = _function_wrapper(fun, args, kwargs)
        self.bounds = asarray(bounds)
        self.n_pop = npop
        self.n_par = self.bounds.shape[0]
        self.bl = tile(self.bounds[:, 0], [npop, 1])
        self.bw = tile(self.bounds[:, 1] - self.bounds[:, 0], [npop, 1])
        self.m = -1 if maximize else 1
        self.pool = pool
        self.args = args

        if self.pool is not None:
            self.map = self.pool.map
        else:
            self.map = map

        self.periodic = []
        self.min_ptp = min_ptp

        self.cmin = cbounds[0]
        self.cmax = cbounds[1]
        self.cbounds = cbounds
        self.fbounds = fbounds

        self.seed = seed
        self.f = f
        self.c = c

        self._population = asarray(self.bl + random([self.n_pop, self.n_par]) * self.bw)
        self._fitness = zeros(npop)
        self._minidx = None

        self._trial_pop = zeros_like(self._population)
        self._trial_fit = zeros_like(self._fitness)

        if vectorize:
            self._eval = self._eval_vfun
        else:
            self._eval = self._eval_sfun

    @property
    def population(self):
        """The parameter vector population"""
        return self._population

    @property
    def minimum_value(self):
        """The best-fit value of the optimized function"""
        return self._fitness[self._minidx]

    @property
    def minimum_location(self):
        """The best-fit solution"""
        return self._population[self._minidx, :]

    @property
    def minimum_index(self):
        """Index of the best-fit solution"""
        return self._minidx

    def optimize(self, ngen):
        """Run the optimizer for ``ngen`` generations"""
        for res in self(ngen):
            pass
        return res

    def __call__(self, ngen=1):
        return self._eval(ngen)

    def _eval_sfun(self, ngen=1):
        """Run DE for a function that takes a single pv as an input and retuns a single value."""
        popc, fitc = self._population, self._fitness
        popt, fitt = self._trial_pop, self._trial_fit

        for ipop in range(self.n_pop):
            fitc[ipop] = self.m * self.minfun(popc[ipop, :])

        for igen in range(ngen):
            f = self.f or uniform(*self.fbounds)
            c = self.c or uniform(*self.cbounds)

            popt = evolve_population(popc, popt, f, c)
            fitt[:] = self.m * array(list(self.map(self.minfun, popt)))

            msk = fitt < fitc
            popc[msk, :] = popt[msk, :]
            fitc[msk] = fitt[msk]

            self._minidx = argmin(fitc)
            if fitc.ptp() < self.min_ptp:
                break

            yield popc[self._minidx, :], fitc[self._minidx]

    def _eval_vfun(self, ngen=1):
        """Run DE for a function that takes the whole population as an input and retuns a value for each pv."""
        popc, fitc = self._population, self._fitness
        popt, fitt = self._trial_pop, self._trial_fit

        fitc[:] = self.m * self.minfun(self._population)

        for igen in range(ngen):
            f = self.f or uniform(*self.fbounds)
            c = self.c or uniform(*self.cbounds)

            popt = evolve_population(popc, popt, f, c)
            fitt[:] = self.m * self.minfun(popt)

            msk = fitt < fitc
            popc[msk, :] = popt[msk, :]
            fitc[msk] = fitt[msk]

            self._minidx = argmin(fitc)
            if fitc.ptp() < self.min_ptp:
                break

            yield popc[self._minidx, :], fitc[self._minidx]


class _function_wrapper(object):
    def __init__(self, f, args=[], kwargs={}):
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def __call__(self, x):
        return self.f(x, *self.args, **self.kwargs)


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Union, List

import astropy.units as u

from pathlib import Path

from astropy.stats import sigma_clipped_stats
from astropy.table import Table
from astropy.time import Time
from astropy.timeseries import TimeSeries
from corner import corner
from matplotlib.pyplot import setp
from matplotlib.pyplot import subplots
from numba import njit
from numpy import zeros, squeeze, ceil, arange, digitize, full, nan, \
    sqrt, percentile, isfinite, floor, argsort, ones_like, atleast_2d, median, ndarray, unique, nanmedian, concatenate
from numpy.random import permutation
from pytransit.utils.tess import read_tess_spoc

from .loglikelihood import CeleriteLogLikelihood
from .lpf import BaseLPF
from .. import TransitModel
from ..orbits import epoch
from ..utils.keplerlc import KeplerLC
from ..utils.misc import fold

try:
    from ldtk import tess
    with_ldtk = True
except ImportError:
    with_ldtk = False

@njit
def downsample_time(time, vals, inttime=1.):
    duration = time.max() - time.min()
    nbins = int(ceil(duration / inttime))
    bins = arange(nbins)
    edges = time[0] + bins * inttime
    bids = digitize(time, edges) - 1
    bt, bv, be = full(nbins, nan), zeros(nbins), zeros(nbins)
    for i, bid in enumerate(bins):
        bmask = bid == bids
        if bmask.sum() > 0:
            bt[i] = time[bmask].mean()
            bv[i] = vals[bmask].mean()
            if bmask.sum() > 2:
                be[i] = vals[bmask].std() / sqrt(bmask.sum())
            else:
                be[i] = nan
    m = isfinite(be)
    return bt[m], bv[m], be[m]


class TESSLPF(BaseLPF):
    bjdrefi = 2457000

    def __init__(self, name: str, datadir: Path = None, tic: int = None, zero_epoch: float = None, period: float = None,
                 nsamples: int = 2, trdur: float = 0.125, bldur: float = 0.3, use_pdc=True,
                 sectors: Optional[Union[List[int], str]] = 'all',
                 split_transits=True, separate_noise=False, tm: TransitModel = None, minpt=10):

        times, fluxes, sectors, wns = read_tess_spoc(tic, datadir, sectors=sectors, use_pdc=use_pdc)
        self.lc = KeplerLC(times, fluxes, sectors, zero_epoch, period, trdur, bldur)

        if split_transits:
            times = self.lc.time_per_transit
            fluxes = self.lc.normalized_flux_per_transit
        else:
            times = concatenate(self.lc.time_per_transit)
            fluxes = concatenate(self.lc.normalized_flux_per_transit)

        tref = floor(concatenate(times).min())

        self.zero_epoch = zero_epoch
        self.period = period
        self.transit_duration = trdur
        self.baseline_duration = bldur

        wnids = arange(len(times)) if separate_noise else None
        BaseLPF.__init__(self, name, ['TESS'], times=times, fluxes=fluxes,
                         nsamples=nsamples, exptimes=0.00139, wnids=wnids, tref=tref, tm=tm)
        self.tm.interpolate = False

    def _init_lnlikelihood(self):
        self._add_lnlikelihood_model(CeleriteLogLikelihood(self))

    def add_ldtk_prior(teff, logg, z):
        if with_ldtk:
            super().add_ldtk_prior(teff, logg, z, passbands=(tess,))
        else:
            raise ImportError('Could not import LDTk, cannot add an LDTk prior.')

    def plot_individual_transits(self, solution: str = 'de', pv: ndarray = None, ncols: int = 2, n_samples: int = 100,
                                 xlim: tuple = None, ylim: tuple = None, axs=None, figsize: tuple = None,
                                 remove_baseline: bool = False):

        solution = solution.lower()
        samples = None

        if pv is None:
            if solution == 'local':
                pv = self._local_minimization.x
            elif solution in ('de', 'global'):
                solution = 'global'
                pv = self.de.minimum_location
            elif solution in ('mcmc', 'mc'):
                solution = 'mcmc'
                samples = self.posterior_samples(derived_parameters=False)
                samples = permutation(samples.values)[:n_samples]
                pv = median(samples, 0)
            else:
                raise NotImplementedError("'solution' should be either 'local', 'global', or 'mcmc'")

        t0 = floor(self.times[0].min())
        nrows = int(ceil(self.nlc / ncols))

        if axs is None:
            fig, axs = subplots(nrows, ncols, figsize=figsize, sharey='all', sharex='all', constrained_layout=True)
        else:
            fig, axs = None, axs

        [ax.autoscale(enable=True, axis='x', tight=True) for ax in axs.flat]

        def baseline(pvp):
            pvp = atleast_2d(pvp)
            bl = zeros((pvp.shape[0], self.ofluxa.size))
            for i, pv in enumerate(pvp):
                bl[i] = self._lnlikelihood_models[0].predict_baseline(pv)
            return bl

        if remove_baseline:
            if solution == 'mcmc':
                fbasel = median(baseline(samples), axis=0)
                fmodel, fmodm, fmodp = percentile(self.transit_model(samples), [50, 0.5, 99.5], axis=0)
            else:
                fbasel = squeeze(baseline(pv))
                fmodel, fmodm, fmodp = squeeze(self.transit_model(pv)), None, None
            fobs = self.ofluxa / fbasel
        else:
            if solution == 'mcmc':
                fbasel = median(baseline(samples), axis=0)
                fmodel, fmodm, fmodp = percentile(self.flux_model(samples), [50, 1, 99], axis=0)
            else:
                fbasel = squeeze(baseline(pv))
                fmodel, fmodm, fmodp = squeeze(self.flux_model(pv)), None, None
            fobs = self.ofluxa

        t0, p = pv[[0, 1]]

        for i, sl in enumerate(self.lcslices):
            ax = axs.flat[i]
            t = self.times[i]
            e = epoch(t.mean(), t0, p)
            tc = t0 + e * p
            tt = 24 * (t - tc)
            ax.plot(tt, fobs[sl], 'k.', alpha=0.2)
            ax.plot(tt, fmodel[sl], 'k')

            if solution == 'mcmc':
                ax.fill_between(tt, fmodm[sl], fmodp[sl], zorder=-100, alpha=0.2, fc='k')

            if not remove_baseline:
                ax.plot(tt, fbasel[sl], 'k--', alpha=0.2)

        setp(axs, xlim=xlim, ylim=ylim)
        setp(axs[-1, :], xlabel='Time - T$_c$ [h]')
        setp(axs[:, 0], ylabel='Normalised flux')
        return fig

    def plot_folded_transit(self, method='de', figsize=(13, 6), ylim=(0.9975, 1.002), xlim=None, binwidth=8,
                            remove_baseline: bool = False):
        if method == 'de':
            pv = self.de.minimum_location
            tc, p = pv[[0, 1]]
        else:
            raise NotImplementedError

        phase = p * fold(self.timea, p, tc, 0.5)
        binwidth = binwidth / 24 / 60
        sids = argsort(phase)

        tm = self.transit_model(pv)

        if remove_baseline:
            gp = self._lnlikelihood_models[0]
            bl = squeeze(gp.predict_baseline(pv))
        else:
            bl = ones_like(self.ofluxa)

        bp, bfo, beo = downsample_time(phase[sids], (self.ofluxa / bl)[sids], binwidth)

        fig, ax = subplots(figsize=figsize)
        ax.plot(phase - 0.5 * p, self.ofluxa / bl, '.', alpha=0.15)
        ax.errorbar(bp - 0.5 * p, bfo, beo, fmt='ko')
        ax.plot(phase[sids] - 0.5 * p, tm[sids], 'k')
        xlim = xlim if xlim is not None else 1.01 * (bp[isfinite(bp)][[0, -1]] - 0.5 * p)
        setp(ax, ylim=ylim, xlim=xlim, xlabel='Time - Tc [d]', ylabel='Normalised flux')
        fig.tight_layout()
        return fig

    def plot_basic_posteriors(self):
        df = self.posterior_samples()
        corner(df['tc p rho b k'.split()],
               labels='Zero epoch, Period, Stellar density, impact parameter, radius ratio'.split(', '))


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Contaminated TESS LPF

This module contains the log posterior function (LPF) for a TESS light curve with possible third-light
contamination from an unresolved source inside the photometry aperture.
"""
from pathlib import Path
from typing import Union

from numpy import repeat, inf, where, newaxis, squeeze, atleast_2d, isfinite, concatenate, zeros
from numpy.random import uniform
from uncertainties import ufloat, UFloat
from ldtk import tess

from ..tesslpf import TESSLPF
from ...param import UniformPrior as UP, NormalPrior as NP, PParameter


class CTESSLPF(TESSLPF):
    """Contaminated TESS LPF

    This class implements a log posterior function for a TESS light curve that allows for unknown flux contamination.
    The amount of flux contamination is not constrained.
    """

    def _init_p_planet(self):
        ps = self.ps
        pk2 = [PParameter('k2_true', 'true_area_ratio', 'A_s', UP(0.10 ** 2, 0.75 ** 2), (0.10 ** 2, 0.75 ** 2)),
               PParameter('k2_app', 'apparent_area_ratio', 'A_s', UP(0.10 ** 2, 0.50 ** 2), (0.10 ** 2, 0.50 ** 2))]
        ps.add_passband_block('k2', 1, 2, pk2)
        self._pid_k2 = repeat(ps.blocks[-1].start, self.npb)
        self._start_k2 = ps.blocks[-1].start
        self._sl_k2 = ps.blocks[-1].slice
        self.add_prior(lambda pv: where(pv[:, 5] < pv[:, 4], 0, -inf))

    def transit_model(self, pv):
        pv = atleast_2d(pv)
        flux = super().transit_model(pv)
        cnt = 1. - pv[:, 5] / pv[:, 4]
        return squeeze(cnt[:, newaxis] + (1. - cnt[:, newaxis]) * flux)

    def create_pv_population(self, npop=50):
        pvp = zeros((0, len(self.ps)))
        npv, i = 0, 0
        while npv < npop and i < 10:
            pvp_trial = self.ps.sample_from_prior(npop)
            pvp_trial[:, 5] = pvp_trial[:, 4]
            cref = uniform(0, 0.99, size=npop)
            pvp_trial[:, 4] = pvp_trial[:, 5] / (1. - cref)
            lnl = self.lnposterior(pvp_trial)
            ids = where(isfinite(lnl))
            pvp = concatenate([pvp, pvp_trial[ids]])
            npv = pvp.shape[0]
            i += 1
        pvp = pvp[:npop]
        return pvp



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import timeit
from typing import Tuple, Callable, Union, List, Optional

from numpy import ndarray, array, squeeze, atleast_2d, atleast_1d, zeros, asarray, linspace, sqrt, pi, ones, log, exp, \
    tile, full, isscalar
from scipy.integrate import trapz

from .ldmodel import LDModel
from .numba.ldmodels import *
from .numba.rrmodel import rrmodel_direct_s, rrmodel_interpolated_s, \
    rrmodel_direct_s_simple, rrmodel_direct_v, \
    rrmodel_interpolated_v

from .numba.rrmodel import create_z_grid, calculate_weights_3d
from .transitmodel import TransitModel

__all__ = ['RoadRunnerModel']


class RoadRunnerModel(TransitModel):
    ldmodels = {'uniform': (ld_uniform, ldi_uniform),
                'linear': (ld_linear, ldi_linear),
                'quadratic': (ld_quadratic, ldi_quadratic),
                'quadratic-tri': (ld_quadratic_tri, ldi_quadratic_tri),
                'nonlinear': ld_nonlinear,
                'general': ld_general,
                'square_root': ld_square_root,
                'logarithmic': ld_logarithmic,
                'exponential': ld_exponential,
                'power-2': ld_power_2,
                'power-2-pm': ld_power_2_pm}

    def __init__(self, ldmodel: Union[str, Callable, Tuple[Callable, Callable]] = 'quadratic',
                 interpolate: bool = False, klims: tuple = (0.005, 0.5), nk: int = 256,
                 nzin: int = 20, nzlimb: int = 20, zcut=0.7, ng: int = 50, parallel: bool = False,
                 small_planet_limit: float = 0.05):
        """The RoadRunner transit model by Parviainen (2020).

        Parameters
        ----------
        interpolate : bool, optional
            Use the interpolation method presented in Parviainen (2015) if true.
        klims : tuple, optional
            Radius ratio limits (kmin, kmax) for the interpolated model.
        nk : int, optional
            Radius ratio grid size for the interpolated model.
        nz : int, optional
            Normalized distance grid size for the interpolated model.
        """
        super().__init__()
        self.interpolate = interpolate
        self.parallel = parallel
        self.is_simple = False
        self.splimit = small_planet_limit

        # Set up the limb darkening model
        # --------------------------------
        if isinstance(ldmodel, str):
            try:
                if isinstance(self.ldmodels[ldmodel], tuple):
                    self.ldmodel = self.ldmodels[ldmodel][0]
                    self.ldmmean = self.ldmodels[ldmodel][1]
                else:
                    self.ldmodel = self.ldmodels[ldmodel]
                    self.ldmmean = None
            except KeyError:
                print(
                    f"Unknown limb darkening model: {ldmodel}. Choose from [{', '.join(self.ldmodels.keys())}] or supply a callable function.")
                raise
        elif isinstance(ldmodel, LDModel):
            self.ldmodel = ldmodel
            self.ldmmean = ldmodel._integrate
        elif callable(ldmodel):
            self.ldmodel = ldmodel
            self.ldmmean = None
        elif isinstance(ldmodel, tuple) and callable(ldmodel[0]) and callable(ldmodel[1]):
            self.ldmodel = ldmodel[0]
            self.ldmmean = ldmodel[1]
        else:
            raise NotImplementedError

        # Set the basic variable
        # ----------------------
        self.klims = klims
        self.nk = nk
        self.ng = ng
        self.nzin = nzin
        self.nzlimb = nzlimb
        self.zcut = zcut

        # Declare the basic arrays
        # ------------------------
        self.ze = None
        self.zm = None
        self.mu = None
        self.dk = None
        self.dg = None
        self.weights = None

        self._m_direct_s = None
        self._m_direct_v = None
        self._m_interp_s = None
        self._m_interp_v = None

        self._ldmu = linspace(1, 0, 200)
        self._ldz = sqrt(1 - self._ldmu ** 2)

        self.init_integration(nzin, nzlimb, zcut, ng, nk)

    def set_data(self, time: Union[ndarray, List],
                 lcids: Optional[Union[ndarray, List]] = None,
                 pbids: Optional[Union[ndarray, List]] = None,
                 nsamples: Optional[Union[ndarray, List]] = None,
                 exptimes: Optional[Union[ndarray, List]] = None,
                 epids: Optional[Union[ndarray, List]] = None) -> None:
        super().set_data(time, lcids, pbids, nsamples, exptimes, epids)
        self.set_methods()

    def set_methods(self):
        if self.npb == 1 and all(self.nsamples == 1) and all(self.epids == 0):
            self.is_simple = True
        else:
            self.is_simple = False

    def init_integration(self, nzin, nzlimb, zcut, ng, nk=None):
        self.nk = nk
        self.ng = ng
        self.nzin = nzin
        self.nzlimb = nzlimb
        self.zcut = zcut
        self.ze, self.zm = create_z_grid(zcut, nzin, nzlimb)
        self.mu = sqrt(1 - self.zm ** 2)
        if self.interpolate:
            self.dk, self.dg, self.weights = calculate_weights_3d(nk, self.klims[0], self.klims[1], self.ze, ng)


    def evaluate(self, k: Union[float, ndarray], ldc: Union[ndarray, List], t0: Union[float, ndarray],
                 p: Union[float, ndarray],
                 a: Union[float, ndarray], i: Union[float, ndarray], e: Optional[Union[float, ndarray]] = None,
                 w: Optional[Union[float, ndarray]] = None, copy: bool = True) -> ndarray:
        """Evaluate the transit model for a set of scalar or vector parameters.

        Parameters
        ----------
        k
            Radius ratio(s) either as a single float, 1D vector, or 2D array.
        ldc
            Limb darkening coefficients as a 1D or 2D array.
        t0
            Transit center(s) as a float or a 1D vector.
        p
            Orbital period(s) as a float or a 1D vector.
        a
            Orbital semi-major axis (axes) divided by the stellar radius as a float or a 1D vector.
        i
            Orbital inclination(s) as a float or a 1D vector.
        e : optional
            Orbital eccentricity as a float or a 1D vector.
        w : optional
            Argument of periastron as a float or a 1D vector.

        Notes
        -----
        The model can be evaluated either for one set of parameters or for many sets of parameters simultaneously. In
        the first case, the orbital parameters should all be given as floats. In the second case, the orbital parameters
        should be given as a 1D array-like.

        Returns
        -------
        ndarray
            Modelled flux either as a 1D or 2D ndarray.
        """

        # Scalar parameters branch
        # ------------------------
        if isscalar(p):
            if e is None:
                e, w = 0.0, 0.0
            return self.evaluate_ps(k, ldc, t0, p, a, i, e, w, copy)

        # Parameter population branch
        # ---------------------------
        else:
            k, t0, p, a, i = asarray(k), asarray(t0), asarray(p), asarray(a), asarray(i)

            if k.ndim == 1:
                k = k.reshape((k.size, 1))

            if t0.ndim == 1:
                t0 = t0.reshape((t0.size, 1))

            npv = p.size
            if e is None:
                e, w = zeros(npv), zeros(npv)

            if isinstance(self.ldmodel, LDModel):
                ldp, istar = self.ldmodel(self.mu, ldc)
            else:
                ldp = evaluate_ld(self.ldmodel, self.mu, ldc)

                if self.ldmmean is not None:
                    istar = evaluate_ldi(self.ldmmean, ldc)
                else:
                    istar = zeros((npv, self.npb))
                    ldpi = evaluate_ld(self.ldmodel, self._ldmu, ldc)
                    for ipv in range(npv):
                        for ipb in range(self.npb):
                            istar[ipv, ipb] = 2 * pi * trapz(self._ldz * ldpi[ipv, ipb], self._ldz)

            if self.interpolate:
                flux = rrmodel_interpolated_v(self.time, k, t0, p, a, i, e, w, ldp, istar, self.weights, self.dk,
                                       self.klims[0], self.dg, self.lcids, self.pbids, self.epids, self.nsamples,
                                       self.exptimes, self.npb, self.parallel)
            else:
                flux = rrmodel_direct_v(self.time, k, t0, p, a, i, e, w, ldp, istar, self.ze, self.ng,
                                       self.lcids, self.pbids, self.epids, self.nsamples,
                                       self.exptimes, self.npb, self.parallel)
        return squeeze(flux)

    def evaluate_ps(self, k: Union[float, ndarray], ldc: ndarray, t0: Union[float, ndarray], p: float, a: float, i: float,
                    e: float = 0.0, w: float = 0.0, copy: bool = True) -> ndarray:
        """Evaluate the transit model for a set of scalar parameters.

        Parameters
        ----------
        k : array-like
            Radius ratio(s) either as a single float or an 1D array.
        ldc : array-like
            Limb darkening coefficients as a 1D array.
        t0 : float
            Transit center as a float.
        p : float
            Orbital period as a float.
        a : float
            Orbital semi-major axis divided by the stellar radius as a float.
        i : float
            Orbital inclination(s) as a float.
        e : float, optional
            Orbital eccentricity as a float.
        w : float, optional
            Argument of periastron as a float.

        Notes
        -----
        This version of the `evaluate` method is optimized for calculating a single transit model (such as when using a
        local optimizer). If you want to evaluate the model for a large number of parameters simultaneously, use either
        `evaluate` or `evaluate_pv`.

        Returns
        -------
        ndarray
            Modelled flux as a 1D ndarray.
        """

        k = asarray(k)
        ldc = asarray(ldc)
        t0 = asarray(t0)

        if isinstance(self.ldmodel, LDModel):
            ldp, istar = self.ldmodel(self.mu, ldc)
        else:
            ldp = evaluate_ld(self.ldmodel, self.mu, ldc)
            if self.ldmmean is not None:
                istar = evaluate_ldi(self.ldmmean, ldc)
            else:
                istar = zeros((1,self.npb))
                if ldc.ndim == 1:
                    ldc = ldc.reshape((1, 1, -1))
                elif ldc.ndim == 2:
                    ldc = ldc.reshape((1, ldc.shape[1], -1))

                for ipb in range(self.npb):
                    istar[0,ipb] = 2 * pi * trapz(self._ldz * self.ldmodel(self._ldmu, ldc[0,ipb]), self._ldz)

        if self.interpolate:
            flux = rrmodel_interpolated_s(self.time, k, t0, p, a, i, e, w, ldp, istar, self.weights, self.zm,
                                    self.dk, self.klims[0], self.dg, self.splimit,
                                    self.lcids, self.pbids, self.epids, self.nsamples, self.exptimes, self.parallel)
        else:
            flux = rrmodel_direct_s(self.time, k, t0, p, a, i, e, w, ldp, istar, self.ze, self.zm, self.ng, self.splimit,
                                    self.lcids, self.pbids, self.epids, self.nsamples, self.exptimes, self.parallel)

        return squeeze(flux)



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2021  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import njit, prange
from numpy import atleast_1d, asarray, zeros, pi, cos, sin, fabs


@njit(parallel=True)
def ellipsoidal_variation(a, t0, p, l, x_is_time, x):
    a = atleast_1d(asarray(a))
    t0 = atleast_1d(asarray(t0))
    p = atleast_1d(asarray(p))
    l = atleast_1d(asarray(l))
    npv = a.size
    npt = x.size
    flux = zeros((npv, npt))
    for i in prange(npv):
        if x_is_time:
            for j in range(npt):
                f = 2.0 * pi * (x[j] - t0[i]) / p[i]
                flux[i, j] = -a[i] * cos(2.0 * (f-l[i]))
        else:
            for j in range(npt):
                flux[i, j] = -a[i] * cos(2.0 * (x[j]-l[i]))
    return flux


@njit(parallel=True)
def doppler_boosting(a, t0, p, x_is_time, x):
    a = atleast_1d(asarray(a))
    t0 = atleast_1d(asarray(t0))
    p = atleast_1d(asarray(p))
    npv = a.size
    npt = x.size
    flux = zeros((npv, npt))
    for i in range(npv):
        if x_is_time:
            for j in range(npt):
                f = 2.0 * pi * (x[j] - t0[i]) / p[i]
                flux[i, j] = a[i] * sin(f)
        else:
            for j in range(npt):
                flux[i, j] = a[i] * sin(x[j])
    return flux


@njit(parallel=True)
def ev_and_db(aev, adb, t0, p, l, x_is_time, x):
    aev = atleast_1d(asarray(aev))
    adb = atleast_1d(asarray(adb))
    t0 = atleast_1d(asarray(t0))
    p = atleast_1d(asarray(p))
    l = atleast_1d(asarray(l))
    npv = aev.size
    npt = x.size
    flux = zeros((npv, npt))
    for i in prange(npv):
        if x_is_time:
            for j in range(npt):
                f = 2.0 * pi * (x[j] - t0[i]) / p[i]
                flux[i, j] = -aev[i] * cos(2.0 * (f-l[i])) + adb[i] * sin(f)
        else:
            for j in range(npt):
                flux[i, j] = -aev[i] * cos(2.0 * (x[j]-l[i])) + adb[i] * sin(x[j])
    return flux


@njit(parallel=True)
def emission(aratio, fr_night, fr_day, offset, t0, p, x_is_time, x):
    aratio = atleast_1d(asarray(aratio))
    fr_night = atleast_1d(asarray(fr_night))
    fr_day = atleast_1d(asarray(fr_day))
    offset = atleast_1d(asarray(offset))
    t0 = atleast_1d(asarray(t0))
    p = atleast_1d(asarray(p))
    npv = aratio.size
    npt = x.size
    flux = zeros((npv, npt))
    for i in prange(npv):
        if x_is_time:
            for j in range(npt):
                f = 2.0 * pi * (x[j] - t0[i]) / p[i]
                flux[i, j] = aratio[i] * (fr_night[i] + (fr_day[i] - fr_night[i]) * 0.5 * (1.0 - cos(f + offset[i])))
        else:
            for j in range(npt):
                flux[i, j] = aratio[i] * (fr_night[i] + (fr_day[i] - fr_night[i]) * 0.5 * (1.0 - cos(x[j] + offset[i])))
    return flux


@njit(parallel=True)
def lambert_phase_function(a, ar, ga, t0, p, x_is_time, x):
    a = atleast_1d(asarray(a))  # Scaled semi-major axis
    ar = atleast_1d(asarray(ar))  # Area ratio
    ga = atleast_1d(asarray(ga))  # geometric albedo
    t0 = atleast_1d(asarray(t0))
    p = atleast_1d(asarray(p))
    npv = a.size
    npt = x.size
    flux = zeros((npv, npt))
    for i in prange(npv):
        amp = ar[i] * ga[i] / a[i]**2
        if x_is_time:
            for j in range(npt):
                f = 2.0 * pi * (x[j] - t0[i]) / p[i]
                alpha = fabs(f % (2.0 * pi) - pi)
                flux[i, j] = amp * (sin(alpha) + (pi - alpha) * cos(alpha)) / pi
        else:
            for j in range(npt):
                alpha = fabs(x[j] % (2.0 * pi) - pi)
                flux[i, j] = amp * (sin(alpha) + (pi - alpha) * cos(alpha)) / pi
    return flux


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Union, Optional, List

from numpy import ndarray, array, squeeze, atleast_2d, atleast_1d, zeros, asarray

from .numba.general_nb import init_arrays, general_model_s, general_model_v, general_model_pv

from .transitmodel import TransitModel

__all__ = ['GeneralModel']

class GeneralModel(TransitModel):
    """Transit model with general limb darkening (Giménez, A&A 450, 1231–1237, 2006).

    Transit model with general limb darkening (Giménez, A&A 450, 1231–1237, 2006) with optimizations described in
    Parviainen (MNRAS 450, 3233–3238, 2015).

    The general limb darkening law is

    .. math:: I(\mu) = I(1) (1 - \sum_{n=1}^N u_n(1-\mu^n) )

    """

    def __init__(self, npol: int = 50, nldc: int = 2, mode: int = 0):
        if npol < 1:
            raise ValueError(f"`npol` has to be positive, now {npol}")
        if nldc <= 0:
            raise ValueError(f"`nldc` cannot be negative, now {nldc}")
        if mode not in (0,1):
            raise ValueError("`mode` has to be either 0 (normal) or 1 (transmission spectroscopy).")

        super().__init__()
        self.npol = npol
        self.nldc = nldc
        self.mode = mode
        self._anm, self._avl, self._ajd, self._aje = init_arrays(npol, nldc)


    def evaluate(self, k: Union[float, ndarray], ldc: Union[ndarray, List], t0: Union[float, ndarray], p: Union[float, ndarray],
                 a: Union[float, ndarray], i: Union[float, ndarray], e: Optional[Union[float, ndarray]] = None,
                 w: Optional[Union[float, ndarray]] = None, copy: bool = True) -> ndarray:
        """Evaluate the transit model for a set of scalar or vector parameters.

        Parameters
        ----------
        k
            Radius ratio(s) either as a single float, 1D vector, or 2D array.
        ldc
            Limb darkening coefficients as a 1D or 2D array.
        t0
            Transit center(s) as a float or a 1D vector.
        p
            Orbital period(s) as a float or a 1D vector.
        a
            Orbital semi-major axis (axes) divided by the stellar radius as a float or a 1D vector.
        i
            Orbital inclination(s) as a float or a 1D vector.
        e : optional
            Orbital eccentricity as a float or a 1D vector.
        w : optional
            Argument of periastron as a float or a 1D vector.

        Notes
        -----
        The model can be evaluated either for one set of parameters or for many sets of parameters simultaneously. In
        the first case, the orbital parameters should all be given as floats. In the second case, the orbital parameters
        should be given as a 1D array-like.

        Returns
        -------
        ndarray
            Modelled flux either as a 1D or 2D ndarray.
        """

        # Scalar parameters branch
        # ------------------------
        if isinstance(t0, float):
            if e is None:
                e, w = 0.0, 0.0
            return self.evaluate_ps(k, ldc, t0, p, a, i, e, w, copy)

        # Parameter population branch
        # ---------------------------
        else:
            ldc = atleast_2d(ldc)
            k, t0, p, a, i = asarray(k), asarray(t0), asarray(p), asarray(a), asarray(i)
            if k.ndim == 1:
                k = k.reshape((k.size,1))
            npv = t0.size
            if e is None:
                e, w = zeros(npv), zeros(npv)
            flux = general_model_v(self.time, k, t0, p, a, i, e, w, ldc, self.mode,
                                   self.lcids, self.pbids, self.nsamples, self.exptimes,
                                   self.npb, self.npol, self.nldc,
                                   self._es, self._ms, self._tae,
                                   self._anm, self._avl, self._ajd, self._aje)
        return squeeze(flux)

    def evaluate_ps(self, k: Union[float, ndarray], ldc: ndarray, t0: float, p: float, a: float, i: float,
                    e: float = 0.0, w: float = 0.0, copy: bool = True) -> ndarray:
        """Evaluate the transit model for a set of scalar parameters.

        Parameters
        ----------
        k : array-like
            Radius ratio(s) either as a single float or an 1D array.
        ldc : array-like
            Limb darkening coefficients as a 1D array.
        t0 : float
            Transit center as a float.
        p : float
            Orbital period as a float.
        a : float
            Orbital semi-major axis divided by the stellar radius as a float.
        i : float
            Orbital inclination(s) as a float.
        e : float, optional
            Orbital eccentricity as a float.
        w : float, optional
            Argument of periastron as a float.

        Notes
        -----
        This version of the `evaluate` method is optimized for calculating a single transit model (such as when using a
        local optimizer). If you want to evaluate the model for a large number of parameters simultaneously, use either
        `evaluate` or `evaluate_pv`.

        Returns
        -------
        ndarray
            Modelled flux as a 1D ndarray.
        """

        ldc = asarray(ldc)
        k = asarray(k)

        if self.time is None:
            raise ValueError("Need to set the data before calling the transit model.")
        if ldc.size != self.nldc * self.npb:
            raise ValueError("The quadratic model needs two limb darkening coefficients per passband")

        flux = general_model_s(self.time, k, t0, p, a, i, e, w, ldc, self.mode,
                               self.lcids, self.pbids, self.nsamples, self.exptimes,
                               self.npb, self.npol, self.nldc,
                               self._es, self._ms, self._tae,
                               self._anm, self._avl, self._ajd, self._aje)
        return squeeze(flux)

    def evaluate_pv(self, pvp: ndarray, ldc: ndarray, copy: bool = True) -> ndarray:
        """Evaluate the transit model for a 2D parameter array.

        Parameters
        ----------
        pvp: ndarray
            Parameter array with a shape `(npv, npar)` where `npv` is the number of parameter vectors, and each row
            contains a set of parameters `[k, t0, p, a, i, e, w]`. The radius ratios can also be given per passband,
            in which case the row should be structured as `[k_0, k_1, k_2, ..., k_npb, t0, p, a, i, e, w]`.
        ldc: ndarray
            Limb darkening coefficient array with shape `(npv, 2*npb)`, where `npv` is the number of parameter vectors
            and `npb` is the number of passbands.

        Notes
        -----
        This version of the `evaluate` method is optimized for calculating several models in parallel, such as when
        using *emcee* for MCMC sampling.

        Returns
        -------
        ndarray
            Modelled flux either as a 1D or 2D ndarray.
        """

        ldc = asarray(ldc)
        pvp = asarray(pvp)

        if self.time is None:
            raise ValueError("Need to set the data before calling the transit model.")

        flux = general_model_pv(self.time, pvp, ldc, self.mode,
                               self.lcids, self.pbids, self.nsamples, self.exptimes,
                               self.npb, self.npol, self.nldc,
                               self._es, self._ms, self._tae,
                               self._anm, self._avl, self._ajd, self._aje)
        return squeeze(flux)

    def to_opencl(self):
        """Creates an OpenCL clone (`QuadraticModelCL`) of the transit model.

        Returns
        -------
        QuadraticModelCL
        """
        from .ma_quadratic_cl import QuadraticModelCL
        tm = QuadraticModelCL(klims=self.klims)
        tm.set_data(self.time, self.lcids, self.pbids, self.nsamples, self.exptimes)
        return tm


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#

from numpy import asarray, unique, zeros, inf, squeeze, zeros_like, isfinite, log10, diff, sqrt
from astropy.stats import mad_std

try:
    from celerite import GP
    from celerite.terms import Matern32Term
    with_celerite = True
except ImportError:
    with_celerite = False

from ...param import  LParameter, NormalPrior as NP, UniformPrior as UP


class MultiCeleriteLogLikelihood:
    def __init__(self, lpf, name: str = 'gp', lcids=None, fixed_hps=None):
        if not with_celerite:
            raise ImportError("MultiCeleriteLogLikelihood requires celerite.")

        self.name = name
        self.lpf = lpf

        if fixed_hps is None:
            self.free = True
        else:
            self.hps = asarray(fixed_hps)
            self.free = False

        if lpf.lcids is None:
            raise ValueError('The LPF data needs to be initialised before initialising CeleriteLogLikelihood.')
        self.lcids = lcids if lcids is not None else unique(lpf.lcids)
        self.lcslices = [lpf.lcslices[i] for i in self.lcids]
        self.nlc = len(self.lcslices)

        self.times, self.fluxes, self.wns = [], [], []
        for i in self.lcids:
            self.times.append(lpf.times[i])
            self.fluxes.append(lpf.fluxes[i])
            self.wns.append(diff(lpf.fluxes[i]).std()/sqrt(2))

        self.gps = [GP(Matern32Term(0, 0)) for i in range(self.nlc)]

        if self.free:
            self.init_parameters()
        else:
            self.compute_gp(None, force=True, hps=self.hps)

    def init_parameters(self):
        name = self.name
        pgp = [LParameter(f'{name}_ln_out', f'{name} ln output scale', '', NP(-6.5, 0.5), bounds=(-inf, inf)),
               LParameter(f'{name}_ln_in', f'{name} ln input scale', '', UP(-8, 8), bounds=(-inf, inf))]
        self.lpf.ps.thaw()
        self.lpf.ps.add_global_block(self.name, pgp)
        self.lpf.ps.freeze()
        self.pv_slice = self.lpf.ps.blocks[-1].slice
        self.pv_start = self.lpf.ps.blocks[-1].start
        setattr(self.lpf, f"_sl_{name}", self.pv_slice)
        setattr(self.lpf, f"_start_{name}", self.pv_start)

    def compute_gp(self, pv, force: bool = False, hps=None):
        if self.free or force:
            parameters = pv[self.pv_slice] if hps is None else hps
            for i, gp in enumerate(self.gps):
                gp.set_parameter_vector(parameters)
                gp.compute(self.times[i], yerr=self.wns[i])

    def compute_gp_lnlikelihood(self, pv, model):
        self.compute_gp(pv)
        lnlike = 0.0
        for i, gp in enumerate(self.gps):
            lnlike += gp.log_likelihood(self.fluxes[i] - model[self.lcslices[i]])
        return lnlike

    def predict_baseline(self, pv):
        self.compute_gp(pv)
        residuals = self.lpf.ofluxa - squeeze(self.lpf.transit_model(pv))
        bl = zeros_like(self.lpf.timea)
        for i, gp in enumerate(self.gps):
            sl = self.lcslices[i]
            bl[sl] = gp.predict(residuals[sl], self.times[i], return_cov=False)
        return 1. + bl

    def __call__(self, pvp, model):
        if pvp.ndim == 1:
            lnlike = self.compute_gp_lnlikelihood(pvp, model)
        else:
            lnlike = zeros(pvp.shape[0])
            for ipv, pv in enumerate(pvp):
                if all(isfinite(model[ipv])):
                    lnlike[ipv] = self.compute_gp_lnlikelihood(pv, model[ipv])
                else:
                    lnlike[ipv] = -inf
        return lnlike


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2021  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import prange, njit
from numpy import atleast_2d, zeros, log, pi, asarray, unique, array, inf, arange, exp

from pytransit.param import  LParameter, NormalPrior as NP, UniformPrior as UP

@njit(fastmath=True)
def lnlike_logistic(o, m, e, slices, nids):
    m = atleast_2d(m)
    slices = atleast_2d(slices)

    npv = m.shape[0]
    nsl = slices.shape[0]
    lnl = zeros(npv)
    for i in prange(npv):
        for isl in range(nsl):
            _e = e[i, nids[isl]]
            for j in range(slices[isl, 0], slices[isl, 1]):
                b = exp(-(o[j]-m[i, j])/ _e)
                lnl[i] += log(b / (_e * (1.+b)**2))
    return lnl

class LogisticLogLikelihood:
    def __init__(self, lpf, name: str = 'wn', noise_ids=None, sigma=None):
        self.name = name
        self.lpf = lpf

        if sigma is None:
            self.free = True
        else:
            self.sigma = asarray(sigma)
            self.free = False

        if lpf.noise_ids is None:
            raise ValueError('The LPF data needs to be initialised before initialising LogisticLogLikelihood.')

        self.global_noise_ids = noise_ids if noise_ids is not None else unique(lpf.noise_ids)
        self.mapping = {g:l for g,l in zip(self.global_noise_ids, arange(self.global_noise_ids.size))}

        slices, lnids = [], []
        for nid, sl in zip(lpf.noise_ids, lpf.lcslices):
            if nid in self.global_noise_ids:
                slices.append([sl.start, sl.stop])
                lnids.append(self.mapping[nid])
        self.lcslices = array(slices)
        self.local_pv_noise_ids = array(lnids)

        self.times = lpf.timea
        self.fluxes = lpf.ofluxa

        if self.free:
            self.init_parameters()

    def init_parameters(self):
        name = self.name
        pgp = [LParameter(f'{name}_loge_{i}', f'{name} log10 sigma {i}', '', UP(-4, 0), bounds=(-inf, inf)) for i in self.global_noise_ids]
        self.lpf.ps.thaw()
        self.lpf.ps.add_global_block(self.name, pgp)
        self.lpf.ps.freeze()
        self.pv_slice = self.lpf.ps.blocks[-1].slice
        self.pv_start = self.lpf.ps.blocks[-1].start
        setattr(self.lpf, f"_sl_{name}", self.pv_slice)
        setattr(self.lpf, f"_start_{name}", self.pv_start)

    def __call__(self, pvp, model):
        e = 10 ** atleast_2d(pvp)[:, self.pv_slice]
        return lnlike_logistic(self.fluxes, model, e, self.lcslices, self.local_pv_noise_ids)


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numpy.random.mtrand import normal, uniform
from pandas import DataFrame
from astropy.units import Unit, R_jup, R_sun, AU
from numpy import sqrt

from ..utils.phasecurves import equilibrium_temperature
from ..orbits import as_from_rhop, i_from_ba, d_from_pkaiews


def derive_qois(data: DataFrame, rstar: tuple = None, teff: tuple = None, distance_unit: Unit = R_jup):
    df = data.copy()
    ns = df.shape[0]

    df['period'] = period = df.p.values if 'p' in df else df.pr.values

    if 'k2_true' in df:
        df['k_true'] = sqrt(df.k2_true)
    if 'k2_app' in df:
        df['k_app'] = sqrt(df.k2_app)

    if 'k2_true' in df and 'k2_app' in df:
        df['cnt'] = 1. - df.k2_app / df.k2_true

    if 'g' in df:
        if 'k' in df:
            df['b'] = df.g * (1 + df.k)
        elif 'k_true' in df:
            df['b'] = df.g * (1 + df.k_true)

    df['a'] = as_from_rhop(df.rho.values, period)
    df['inc'] = i_from_ba(df.b.values, df.a.values)
    df['t14'] = d_from_pkaiews(period, df.k_true.values, df.a.values, df.inc.values, 0.0, 0.0, 1)
    df['t14_h'] = 24 * df.t14

    if rstar is not None:
        from astropy.units import R_sun
        rstar_d = (normal(*rstar, size=ns) * R_sun).to(distance_unit).value
        df['r_app'] = df.k_app.values * rstar_d
        df['r_true'] = df.k_true.values * rstar_d
        df['a_au'] = df.a * (rstar_d * distance_unit).to(AU)

    if teff is not None:
        df['teq_p'] = equilibrium_temperature(normal(*teff, size=ns), df.a, uniform(0.25, 0.50, ns), uniform(0, 0.4, ns))
    return df